/*

  Copyright (C) 2000, 2001 Silicon Graphics, Inc.  All Rights Reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of version 2 of the GNU General Public License as
  published by the Free Software Foundation.

  This program is distributed in the hope that it would be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

  Further, this software is distributed without any warranty that it is
  free of the rightful claim of any third person regarding infringement 
  or the like.  Any license provided herein, whether implied or 
  otherwise, applies only to this software file.  Patent licenses, if 
  any, provided herein do not apply to combinations of this program with 
  other software, or any other product whatsoever.  

  You should have received a copy of the GNU General Public License along
  with this program; if not, write the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston MA 02111-1307, USA.

  Contact information:  Silicon Graphics, Inc., 1600 Amphitheatre Pky,
  Mountain View, CA 94043, or:

  http://www.sgi.com

  For further information regarding this notice, see:

  http://oss.sgi.com/projects/GenInfo/NoticeExplan

*/


#ifndef opcode_gen_core_INCLUDED
#define opcode_gen_core_INCLUDED

#ifndef mtypes_INCLUDED
#include "mtypes.h"
#endif

#ifndef ir_a2b_util_INCLUDED
#include "ir_a2b_util.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* annotation categories */

typedef enum {
  OPERATOR_MAPCAT_HDR    = 0,
  OPERATOR_MAPCAT_SCF    = 1,
  OPERATOR_MAPCAT_LDST   = 2,
  OPERATOR_MAPCAT_PRAGMA = 3,
  OPERATOR_MAPCAT_OSTMT  = 4,
  OPERATOR_MAPCAT_OEXP   = 5,
  OPERATOR_MAPCAT_ARRAY  = 6,
  OPERATOR_MAPCAT_CALL   = 7
} OPERATOR_MAPCAT;

#define WN_MAP_CATEGORIES 8


/* Flag property bits, field _flag */

#define OPERATOR_PROPERTY_scf			0x00000001
#define OPERATOR_PROPERTY_stmt			0x00000002
#define OPERATOR_PROPERTY_expression		0x00000004
#define OPERATOR_PROPERTY_leaf			0x00000008
#define OPERATOR_PROPERTY_store			0x00000010
#define OPERATOR_PROPERTY_load			0x00000020
#define OPERATOR_PROPERTY_call			0x00000040
#define OPERATOR_PROPERTY_compare		0x00000080
#define OPERATOR_PROPERTY_non_scf		0x00000100
#define OPERATOR_PROPERTY_boolean		0x00000200
#define OPERATOR_PROPERTY_endsbb		0x00000400
#define OPERATOR_PROPERTY_comp_unit_if		0x00000800
#define OPERATOR_PROPERTY_not_executable	0x00001000
#define OPERATOR_PROPERTY_prefetch		0x00002000
#define OPERATOR_PROPERTY_next_prev		0x00004000
#define OPERATOR_PROPERTY_sym			0x00008000
#define OPERATOR_PROPERTY_label			0x00010000
#define OPERATOR_PROPERTY_num_entries		0x00020000
#define OPERATOR_PROPERTY_offset		0x00040000
#define OPERATOR_PROPERTY_2offsets		0x00080000
#define OPERATOR_PROPERTY_bits			0x00100000
#define OPERATOR_PROPERTY_ndim			0x00200000
#define OPERATOR_PROPERTY_esize			0x00400000
#define OPERATOR_PROPERTY_value			0x00800000
#define OPERATOR_PROPERTY_flags			0x01000000
#define OPERATOR_PROPERTY_inumber		0x02000000
#define OPERATOR_PROPERTY_1ty			0x04000000
#define OPERATOR_PROPERTY_2ty			0x08000000
#define OPERATOR_PROPERTY_ereg_supp		0x10000000
#define OPERATOR_PROPERTY_barrier		0x20000000
#define OPERATOR_PROPERTY_last_label		0x40000000
#define OPERATOR_PROPERTY_field_id		0x80000000

/* Definition of type OPERATOR */

#define OPR_DIVPART OPR_LOWPART
#define OPR_REMPART OPR_HIGHPART
#define OPR_COMPLEX OPR_PAIR
#define OPR_REALPART OPR_FIRSTPART
#define OPR_IMAGPART OPR_SECONDPART

typedef enum {
  OPERATOR_UNKNOWN = 0,
  OPERATOR_FIRST = 1,
  OPR_ABS = 1,
  OPR_ADD = 2,
  OPR_AGOTO = 3,
  OPR_ALTENTRY = 4,
  OPR_ARRAY = 5,
  OPR_ARRAYEXP = 6,
  OPR_ARRSECTION = 7,
  OPR_ASHR = 8,
  OPR_ASSERT = 9,
  OPR_BACKWARD_BARRIER = 10,
  OPR_BAND = 11,
  OPR_BIOR = 12,
  OPR_BLOCK = 13,
  OPR_BNOR = 14,
  OPR_BNOT = 15,
  OPR_BXOR = 16,
  OPR_CALL = 17,
  OPR_CAND = 18,
  OPR_CASEGOTO = 19,
  OPR_CEIL = 20,
  OPR_CIOR = 21,
  OPR_COMMA = 22,
  OPR_COMMENT = 23,
  OPR_COMPGOTO = 24,
  OPR_PAIR = 25,
  OPR_CONST = 26,
  OPR_CSELECT = 27,
  OPR_CVT = 28,
  OPR_CVTL = 29,
  OPR_DIV = 30,
  OPR_DIVREM = 31,
  OPR_DO_LOOP = 32,
  OPR_DO_WHILE = 33,
  OPR_EQ = 34,
  OPR_EVAL = 35,
  OPR_EXC_SCOPE_BEGIN = 36,
  OPR_EXC_SCOPE_END = 37,
  OPR_FALSEBR = 38,
  OPR_FLOOR = 39,
  OPR_FORWARD_BARRIER = 40,
  OPR_FUNC_ENTRY = 41,
  OPR_GE = 42,
  OPR_GOTO = 43,
  OPR_GT = 44,
  OPR_HIGHMPY = 45,
  OPR_HIGHPART = 46,
  OPR_ICALL = 47,
  OPR_IDNAME = 48,
  OPR_IF = 49,
  OPR_ILDA = 50,
  OPR_ILDBITS = 51,
  OPR_ILOAD = 52,
  OPR_ILOADX = 53,
  OPR_SECONDPART = 54,
  OPR_INTCONST = 55,
  OPR_INTRINSIC_CALL = 56,
  OPR_INTRINSIC_OP = 57,
  OPR_IO = 58,
  OPR_IO_ITEM = 59,
  OPR_ISTBITS = 60,
  OPR_ISTORE = 61,
  OPR_ISTOREX = 62,
  OPR_LABEL = 63,
  OPR_LAND = 64,
  OPR_LDA = 65,
  OPR_LDBITS = 66,
  OPR_LDID = 67,
  OPR_LE = 68,
  OPR_LIOR = 69,
  OPR_LNOT = 70,
  OPR_LOOP_INFO = 71,
  OPR_LOWPART = 72,
  OPR_LSHR = 73,
  OPR_LT = 74,
  OPR_MADD = 75,
  OPR_MAX = 76,
  OPR_MAXPART = 77,
  OPR_MIN = 78,
  OPR_MINMAX = 79,
  OPR_MINPART = 80,
  OPR_MLOAD = 81,
  OPR_MOD = 82,
  OPR_MPY = 83,
  OPR_MSTORE = 84,
  OPR_MSUB = 85,
  OPR_NE = 86,
  OPR_NEG = 87,
  OPR_NMADD = 88,
  OPR_NMSUB = 89,
  OPR_OPTPARM = 90,
  OPR_OPT_CHI = 91,
  OPR_OPT_RESERVE2 = 92,
  OPR_PAREN = 93,
  OPR_PARM = 94,
  OPR_PICCALL = 95,
  OPR_PRAGMA = 96,
  OPR_PREFETCH = 97,
  OPR_PREFETCHX = 98,
  OPR_RCOMMA = 99,
  OPR_FIRSTPART = 100,
  OPR_RECIP = 101,
  OPR_REGION = 102,
  OPR_REGION_EXIT = 103,
  OPR_REM = 104,
  OPR_RETURN = 105,
  OPR_RETURN_VAL = 106,
  OPR_RND = 107,
  OPR_RSQRT = 108,
  OPR_SELECT = 109,
  OPR_SHL = 110,
  OPR_SQRT = 111,
  OPR_STBITS = 112,
  OPR_STID = 113,
  OPR_SUB = 114,
  OPR_SWITCH = 115,
  OPR_TAS = 116,
  OPR_TRAP = 117,
  OPR_TRIPLET = 118,
  OPR_TRUEBR = 119,
  OPR_TRUNC = 120,
  OPR_VFCALL = 121,
  OPR_WHERE = 122,
  OPR_WHILE_DO = 123,
  OPR_XGOTO = 124,
  OPR_XMPY = 125,
  OPR_XPRAGMA = 126,
  OPR_AFFIRM = 127,
  OPR_ALLOCA = 128,
  OPR_DEALLOCA = 129,
  OPR_LDMA = 130,
  OPR_ASM_STMT = 131,
  OPR_ASM_EXPR = 132,
  OPR_ASM_INPUT = 133,
  OPR_RROTATE = 134,
  OPR_LDA_LABEL = 135,
  OPR_GOTO_OUTER_BLOCK = 136,
  OPR_EXTRACT_BITS = 137,
  OPR_COMPOSE_BITS = 138,
  OPR_PSTORE = 139,  
  OPR_PSTID = 140,   
  OPR_USE  = 141,    
  OPR_NAMELIST = 142, 
  OPR_IMPLICIT_BND = 143, 
  OPR_SRCTRIPLET = 144, 
  OPR_NULLIFY = 145,
  OPR_INTERFACE = 146,
  OPR_ARRAY_CONSTRUCT = 147,
  OPR_IMPLIED_DO =148,
  OPR_STRCTFLD   =149,

  OPERATOR_LAST = 149
} OPERATOR;

/* Definition of type OPCODE */

#define OPC_I4DIVPART OPC_I4LOWPART
#define OPC_I8DIVPART OPC_I8LOWPART
#define OPC_U4DIVPART OPC_U4LOWPART
#define OPC_U8DIVPART OPC_U8LOWPART
#define OPC_I4REMPART OPC_I4HIGHPART
#define OPC_I8REMPART OPC_I8HIGHPART
#define OPC_U4REMPART OPC_U4HIGHPART
#define OPC_U8REMPART OPC_U8HIGHPART

#define OPC_C4COMPLEX   OPC_C4PAIR
#define OPC_C8COMPLEX   OPC_C8PAIR
#define OPC_CQCOMPLEX   OPC_CQPAIR
#define OPC_C10COMPLEX  OPC_C10PAIR
#define OPC_C16COMPLEX  OPC_C16PAIR
#define OPC_F4REALPART  OPC_F4FIRSTPART
#define OPC_F8REALPART  OPC_F8FIRSTPART
#define OPC_FQREALPART  OPC_FQFIRSTPART
#define OPC_F10REALPART OPC_F10FIRSTPART
#define OPC_F16REALPART OPC_F16FIRSTPART
#define OPC_F4IMAGPART  OPC_F4SECONDPART
#define OPC_F8IMAGPART  OPC_F8SECONDPART
#define OPC_FQIMAGPART  OPC_FQSECONDPART
#define OPC_F10IMAGPART OPC_F10SECONDPART
#define OPC_F16IMAGPART OPC_F16SECONDPART

#define RTYPE(x) (x<<8)
#define DESC(x)  (x<<13)

typedef enum {
  OPCODE_UNKNOWN         = 0,
  OPCODE_FIRST           = 1,
  OPC_I4ABS              = OPR_ABS + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8ABS              = OPR_ABS + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_F4ABS              = OPR_ABS + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8ABS              = OPR_ABS + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQABS              = OPR_ABS + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10ABS             = OPR_ABS + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16ABS             = OPR_ABS + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_I4ADD              = OPR_ADD + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8ADD              = OPR_ADD + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4ADD              = OPR_ADD + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8ADD              = OPR_ADD + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4ADD              = OPR_ADD + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8ADD              = OPR_ADD + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQADD              = OPR_ADD + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10ADD             = OPR_ADD + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16ADD             = OPR_ADD + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4ADD              = OPR_ADD + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8ADD              = OPR_ADD + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQADD              = OPR_ADD + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_A4ADD              = OPR_ADD + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8ADD              = OPR_ADD + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10ADD             = OPR_ADD + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16ADD             = OPR_ADD + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16ADD             = OPR_ADD + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16ADD             = OPR_ADD + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_AGOTO              = OPR_AGOTO + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_ALTENTRY           = OPR_ALTENTRY + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_U4ARRAY            = OPR_ARRAY + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8ARRAY            = OPR_ARRAY + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_A4ARRAY            = OPR_ARRAY + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8ARRAY            = OPR_ARRAY + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_I4ARRAYEXP         = OPR_ARRAYEXP + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8ARRAYEXP         = OPR_ARRAYEXP + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U1ARRAYEXP         = OPR_ARRAYEXP + RTYPE(MTYPE_U1) + DESC(MTYPE_V),
  OPC_U4ARRAYEXP         = OPR_ARRAYEXP + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8ARRAYEXP         = OPR_ARRAYEXP + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4ARRAYEXP         = OPR_ARRAYEXP + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8ARRAYEXP         = OPR_ARRAYEXP + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQARRAYEXP         = OPR_ARRAYEXP + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10ARRAYEXP        = OPR_ARRAYEXP + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16ARRAYEXP        = OPR_ARRAYEXP + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4ARRAYEXP         = OPR_ARRAYEXP + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8ARRAYEXP         = OPR_ARRAYEXP + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQARRAYEXP         = OPR_ARRAYEXP + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MARRAYEXP          = OPR_ARRAYEXP + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_C10ARRAYEXP        = OPR_ARRAYEXP + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16ARRAYEXP        = OPR_ARRAYEXP + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16ARRAYEXP        = OPR_ARRAYEXP + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16ARRAYEXP        = OPR_ARRAYEXP + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_U4ARRSECTION       = OPR_ARRSECTION + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8ARRSECTION       = OPR_ARRSECTION + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_A4ARRSECTION       = OPR_ARRSECTION + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8ARRSECTION       = OPR_ARRSECTION + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_I4ASHR             = OPR_ASHR + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8ASHR             = OPR_ASHR + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4ASHR             = OPR_ASHR + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8ASHR             = OPR_ASHR + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16ASHR            = OPR_ASHR + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16ASHR            = OPR_ASHR + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_ASSERT             = OPR_ASSERT + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_BACKWARD_BARRIER   = OPR_BACKWARD_BARRIER + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4BAND             = OPR_BAND + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8BAND             = OPR_BAND + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4BAND             = OPR_BAND + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8BAND             = OPR_BAND + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16BAND            = OPR_BAND + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16BAND            = OPR_BAND + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4BIOR             = OPR_BIOR + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8BIOR             = OPR_BIOR + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4BIOR             = OPR_BIOR + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8BIOR             = OPR_BIOR + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16BIOR            = OPR_BIOR + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16BIOR            = OPR_BIOR + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_BLOCK              = OPR_BLOCK + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4BNOR             = OPR_BNOR + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8BNOR             = OPR_BNOR + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4BNOR             = OPR_BNOR + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8BNOR             = OPR_BNOR + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16BNOR            = OPR_BNOR + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16BNOR            = OPR_BNOR + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4BNOT             = OPR_BNOT + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8BNOT             = OPR_BNOT + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4BNOT             = OPR_BNOT + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8BNOT             = OPR_BNOT + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16BNOT            = OPR_BNOT + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16BNOT            = OPR_BNOT + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4BXOR             = OPR_BXOR + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8BXOR             = OPR_BXOR + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4BXOR             = OPR_BXOR + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8BXOR             = OPR_BXOR + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16BXOR            = OPR_BXOR + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16BXOR            = OPR_BXOR + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I1CALL             = OPR_CALL + RTYPE(MTYPE_I1) + DESC(MTYPE_V),
  OPC_I2CALL             = OPR_CALL + RTYPE(MTYPE_I2) + DESC(MTYPE_V),
  OPC_I4I4CALL           = OPR_CALL + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4CALL             = OPR_CALL + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8I8CALL           = OPR_CALL + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8CALL             = OPR_CALL + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U1CALL             = OPR_CALL + RTYPE(MTYPE_U1) + DESC(MTYPE_V),
  OPC_U2CALL             = OPR_CALL + RTYPE(MTYPE_U2) + DESC(MTYPE_V),
  OPC_U4U4CALL           = OPR_CALL + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4CALL             = OPR_CALL + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8U8CALL           = OPR_CALL + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8CALL             = OPR_CALL + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4F4CALL           = OPR_CALL + RTYPE(MTYPE_F4) + DESC(MTYPE_F4),
  OPC_F4F8CALL           = OPR_CALL + RTYPE(MTYPE_F4) + DESC(MTYPE_F8),
  OPC_F4CALL             = OPR_CALL + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8F4CALL           = OPR_CALL + RTYPE(MTYPE_F8) + DESC(MTYPE_F4),
  OPC_F8F8CALL           = OPR_CALL + RTYPE(MTYPE_F8) + DESC(MTYPE_F8),
  OPC_F8CALL             = OPR_CALL + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQCALL             = OPR_CALL + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10CALL            = OPR_CALL + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16CALL            = OPR_CALL + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4CALL             = OPR_CALL + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8CALL             = OPR_CALL + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQCALL             = OPR_CALL + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MCALL              = OPR_CALL + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_VCALL              = OPR_CALL + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_A4CALL             = OPR_CALL + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8CALL             = OPR_CALL + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10CALL            = OPR_CALL + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16CALL            = OPR_CALL + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_BCAND              = OPR_CAND + RTYPE(MTYPE_B) + DESC(MTYPE_V),
  OPC_I4CAND             = OPR_CAND + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_CASEGOTO           = OPR_CASEGOTO + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4F4CEIL           = OPR_CEIL + RTYPE(MTYPE_I4) + DESC(MTYPE_F4),
  OPC_I4F8CEIL           = OPR_CEIL + RTYPE(MTYPE_I4) + DESC(MTYPE_F8),
  OPC_I4FQCEIL           = OPR_CEIL + RTYPE(MTYPE_I4) + DESC(MTYPE_FQ),
  OPC_I4F10CEIL          = OPR_CEIL + RTYPE(MTYPE_I4) + DESC(MTYPE_F10),
  OPC_I4F16CEIL          = OPR_CEIL + RTYPE(MTYPE_I4) + DESC(MTYPE_F16),
  OPC_I8F4CEIL           = OPR_CEIL + RTYPE(MTYPE_I8) + DESC(MTYPE_F4),
  OPC_I8F8CEIL           = OPR_CEIL + RTYPE(MTYPE_I8) + DESC(MTYPE_F8),
  OPC_I8FQCEIL           = OPR_CEIL + RTYPE(MTYPE_I8) + DESC(MTYPE_FQ),
  OPC_I8F10CEIL          = OPR_CEIL + RTYPE(MTYPE_I8) + DESC(MTYPE_F10),
  OPC_I8F16CEIL          = OPR_CEIL + RTYPE(MTYPE_I8) + DESC(MTYPE_F16),
  OPC_U4F4CEIL           = OPR_CEIL + RTYPE(MTYPE_U4) + DESC(MTYPE_F4),
  OPC_U4F8CEIL           = OPR_CEIL + RTYPE(MTYPE_U4) + DESC(MTYPE_F8),
  OPC_U4FQCEIL           = OPR_CEIL + RTYPE(MTYPE_U4) + DESC(MTYPE_FQ),
  OPC_U4F10CEIL          = OPR_CEIL + RTYPE(MTYPE_U4) + DESC(MTYPE_F10),
  OPC_U4F16CEIL          = OPR_CEIL + RTYPE(MTYPE_U4) + DESC(MTYPE_F16),
  OPC_U8F4CEIL           = OPR_CEIL + RTYPE(MTYPE_U8) + DESC(MTYPE_F4),
  OPC_U8F8CEIL           = OPR_CEIL + RTYPE(MTYPE_U8) + DESC(MTYPE_F8),
  OPC_U8FQCEIL           = OPR_CEIL + RTYPE(MTYPE_U8) + DESC(MTYPE_FQ),
  OPC_U8F10CEIL          = OPR_CEIL + RTYPE(MTYPE_U8) + DESC(MTYPE_F10),
  OPC_U8F16CEIL          = OPR_CEIL + RTYPE(MTYPE_U8) + DESC(MTYPE_F16),
  OPC_I16F4CEIL          = OPR_CEIL + RTYPE(MTYPE_I16) + DESC(MTYPE_F4),
  OPC_I16F8CEIL          = OPR_CEIL + RTYPE(MTYPE_I16) + DESC(MTYPE_F8),
  OPC_I16FQCEIL          = OPR_CEIL + RTYPE(MTYPE_I16) + DESC(MTYPE_FQ),
  OPC_I16F10CEIL         = OPR_CEIL + RTYPE(MTYPE_I16) + DESC(MTYPE_F10),
  OPC_I16F16CEIL         = OPR_CEIL + RTYPE(MTYPE_I16) + DESC(MTYPE_F16),
  OPC_U16F4CEIL          = OPR_CEIL + RTYPE(MTYPE_U16) + DESC(MTYPE_F4),
  OPC_U16F8CEIL          = OPR_CEIL + RTYPE(MTYPE_U16) + DESC(MTYPE_F8),
  OPC_U16FQCEIL          = OPR_CEIL + RTYPE(MTYPE_U16) + DESC(MTYPE_FQ),
  OPC_U16F10CEIL         = OPR_CEIL + RTYPE(MTYPE_U16) + DESC(MTYPE_F10),
  OPC_U16F16CEIL         = OPR_CEIL + RTYPE(MTYPE_U16) + DESC(MTYPE_F16),
  OPC_BCIOR              = OPR_CIOR + RTYPE(MTYPE_B) + DESC(MTYPE_V),
  OPC_I4CIOR             = OPR_CIOR + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I4COMMA            = OPR_COMMA + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8COMMA            = OPR_COMMA + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4COMMA            = OPR_COMMA + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8COMMA            = OPR_COMMA + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4COMMA            = OPR_COMMA + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8COMMA            = OPR_COMMA + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQCOMMA            = OPR_COMMA + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10COMMA           = OPR_COMMA + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16COMMA           = OPR_COMMA + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4COMMA            = OPR_COMMA + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8COMMA            = OPR_COMMA + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQCOMMA            = OPR_COMMA + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MCOMMA             = OPR_COMMA + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_A4COMMA            = OPR_COMMA + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8COMMA            = OPR_COMMA + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10COMMA           = OPR_COMMA + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16COMMA           = OPR_COMMA + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16COMMA           = OPR_COMMA + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16COMMA           = OPR_COMMA + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_COMMENT            = OPR_COMMENT + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_COMPGOTO           = OPR_COMPGOTO + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_C4PAIR             = OPR_PAIR + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8PAIR             = OPR_PAIR + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQPAIR             = OPR_PAIR + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_C10PAIR            = OPR_PAIR + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16PAIR            = OPR_PAIR + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I4CONST            = OPR_CONST + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8CONST            = OPR_CONST + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4CONST            = OPR_CONST + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8CONST            = OPR_CONST + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4CONST            = OPR_CONST + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8CONST            = OPR_CONST + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQCONST            = OPR_CONST + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10CONST           = OPR_CONST + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16CONST           = OPR_CONST + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4CONST            = OPR_CONST + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8CONST            = OPR_CONST + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQCONST            = OPR_CONST + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_A4CONST            = OPR_CONST + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8CONST            = OPR_CONST + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10CONST           = OPR_CONST + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16CONST           = OPR_CONST + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16CONST           = OPR_CONST + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16CONST           = OPR_CONST + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_BCSELECT           = OPR_CSELECT + RTYPE(MTYPE_B) + DESC(MTYPE_V),
  OPC_I4CSELECT          = OPR_CSELECT + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8CSELECT          = OPR_CSELECT + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4CSELECT          = OPR_CSELECT + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8CSELECT          = OPR_CSELECT + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4CSELECT          = OPR_CSELECT + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8CSELECT          = OPR_CSELECT + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQCSELECT          = OPR_CSELECT + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10CSELECT         = OPR_CSELECT + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16CSELECT         = OPR_CSELECT + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4CSELECT          = OPR_CSELECT + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8CSELECT          = OPR_CSELECT + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQCSELECT          = OPR_CSELECT + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MCSELECT           = OPR_CSELECT + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_VCSELECT           = OPR_CSELECT + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_A4CSELECT          = OPR_CSELECT + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8CSELECT          = OPR_CSELECT + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10CSELECT         = OPR_CSELECT + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16CSELECT         = OPR_CSELECT + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16CSELECT         = OPR_CSELECT + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16CSELECT         = OPR_CSELECT + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4I8CVT            = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_I8),
  OPC_I4U4CVT            = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_U4),
  OPC_I4U8CVT            = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_U8),
  OPC_I4F4CVT            = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_F4),
  OPC_I4F8CVT            = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_F8),
  OPC_I4FQCVT            = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_FQ),
  OPC_I4F10CVT           = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_F10),
  OPC_I4F16CVT           = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_F16),
  OPC_I4A4CVT            = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_A4),
  OPC_I4A8CVT            = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_A8),
  OPC_I4I16CVT           = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_I16),
  OPC_I4U16CVT           = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_U16),
  OPC_I8I4CVT            = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_I4),
  OPC_I8U4CVT            = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_U4),
  OPC_I8U8CVT            = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_U8),
  OPC_I8F4CVT            = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_F4),
  OPC_I8F8CVT            = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_F8),
  OPC_I8FQCVT            = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_FQ),
  OPC_I8F10CVT           = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_F10),
  OPC_I8F16CVT           = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_F16),
  OPC_I8A4CVT            = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_A4),
  OPC_I8A8CVT            = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_A8),
  OPC_I8I16CVT           = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_I16),
  OPC_I8U16CVT           = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_U16),
  OPC_U4I4CVT            = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_I4),
  OPC_U4I8CVT            = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_I8),
  OPC_U4U8CVT            = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_U8),
  OPC_U4F4CVT            = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_F4),
  OPC_U4F8CVT            = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_F8),
  OPC_U4FQCVT            = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_FQ),
  OPC_U4F10CVT           = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_F10),
  OPC_U4F16CVT           = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_F16),
  OPC_U4A4CVT            = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_A4),
  OPC_U4A8CVT            = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_A8),
  OPC_U4I16CVT           = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_I16),
  OPC_U4U16CVT           = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_U16),
  OPC_U8I4CVT            = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_I4),
  OPC_U8I8CVT            = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_I8),
  OPC_U8U4CVT            = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_U4),
  OPC_U8F4CVT            = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_F4),
  OPC_U8F8CVT            = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_F8),
  OPC_U8FQCVT            = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_FQ),
  OPC_U8F10CVT           = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_F10),
  OPC_U8F16CVT           = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_F16),
  OPC_U8A4CVT            = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_A4),
  OPC_U8A8CVT            = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_A8),
  OPC_U8I16CVT           = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_I16),
  OPC_U8U16CVT           = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_U16),
  OPC_F4I4CVT            = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_I4),
  OPC_F4I8CVT            = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_I8),
  OPC_F4U4CVT            = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_U4),
  OPC_F4U8CVT            = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_U8),
  OPC_F4F8CVT            = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_F8),
  OPC_F4FQCVT            = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_FQ),
  OPC_F4F10CVT           = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_F10),
  OPC_F4F16CVT           = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_F16),
  OPC_F4A4CVT            = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_A4),
  OPC_F4A8CVT            = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_A8),
  OPC_F4I16CVT           = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_I16),
  OPC_F4U16CVT           = OPR_CVT + RTYPE(MTYPE_F4) + DESC(MTYPE_U16),
  OPC_F8I4CVT            = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_I4),
  OPC_F8I8CVT            = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_I8),
  OPC_F8U4CVT            = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_U4),
  OPC_F8U8CVT            = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_U8),
  OPC_F8F4CVT            = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_F4),
  OPC_F8FQCVT            = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_FQ),
  OPC_F8F10CVT           = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_F10),
  OPC_F8F16CVT           = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_F16),
  OPC_F8A4CVT            = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_A4),
  OPC_F8A8CVT            = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_A8),
  OPC_F8I16CVT           = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_I16),
  OPC_F8U16CVT           = OPR_CVT + RTYPE(MTYPE_F8) + DESC(MTYPE_U16),
  OPC_FQI4CVT            = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_I4),
  OPC_FQI8CVT            = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_I8),
  OPC_FQU4CVT            = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_U4),
  OPC_FQU8CVT            = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_U8),
  OPC_FQF4CVT            = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_F4),
  OPC_FQF8CVT            = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_F8),
  OPC_FQF10CVT           = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_F10),
  OPC_FQF16CVT           = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_F16),
  OPC_FQA4CVT            = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_A4),
  OPC_FQA8CVT            = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_A8),
  OPC_FQI16CVT           = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_I16),
  OPC_FQU16CVT           = OPR_CVT + RTYPE(MTYPE_FQ) + DESC(MTYPE_U16),
  OPC_F10I4CVT           = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_I4),
  OPC_F10I8CVT           = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_I8),
  OPC_F10U4CVT           = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_U4),
  OPC_F10U8CVT           = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_U8),
  OPC_F10F4CVT           = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_F4),
  OPC_F10F8CVT           = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_F8),
  OPC_F10FQCVT           = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_FQ),
  OPC_F10F16CVT          = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_F16),
  OPC_F10A4CVT           = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_A4),
  OPC_F10A8CVT           = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_A8),
  OPC_F10I16CVT          = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_I16),
  OPC_F10U16CVT          = OPR_CVT + RTYPE(MTYPE_F10) + DESC(MTYPE_U16),
  OPC_F16I4CVT           = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_I4),
  OPC_F16I8CVT           = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_I8),
  OPC_F16U4CVT           = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_U4),
  OPC_F16U8CVT           = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_U8),
  OPC_F16F4CVT           = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_F4),
  OPC_F16F8CVT           = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_F8),
  OPC_F16FQCVT           = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_FQ),
  OPC_F16F10CVT          = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_F10),
  OPC_F16A4CVT           = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_A4),
  OPC_F16A8CVT           = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_A8),
  OPC_F16I16CVT          = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_I16),
  OPC_F16U16CVT          = OPR_CVT + RTYPE(MTYPE_F16) + DESC(MTYPE_U16),
  OPC_A4I4CVT            = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_I4),
  OPC_A4I8CVT            = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_I8),
  OPC_A4U4CVT            = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_U4),
  OPC_A4U8CVT            = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_U8),
  OPC_A4F4CVT            = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_F4),
  OPC_A4F8CVT            = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_F8),
  OPC_A4FQCVT            = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_FQ),
  OPC_A4F10CVT           = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_F10),
  OPC_A4F16CVT           = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_F16),
  OPC_A4A8CVT            = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_A8),
  OPC_A4I16CVT           = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_I16),
  OPC_A4U16CVT           = OPR_CVT + RTYPE(MTYPE_A4) + DESC(MTYPE_U16),
  OPC_A8I4CVT            = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_I4),
  OPC_A8I8CVT            = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_I8),
  OPC_A8U4CVT            = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_U4),
  OPC_A8U8CVT            = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_U8),
  OPC_A8F4CVT            = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_F4),
  OPC_A8F8CVT            = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_F8),
  OPC_A8FQCVT            = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_FQ),
  OPC_A8F10CVT           = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_F10),
  OPC_A8F16CVT           = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_F16),
  OPC_A8A4CVT            = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_A4),
  OPC_A8I16CVT           = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_I16),
  OPC_A8U16CVT           = OPR_CVT + RTYPE(MTYPE_A8) + DESC(MTYPE_U16),
  OPC_I16I4CVT           = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_I4),
  OPC_I16I8CVT           = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_I8),
  OPC_I16U4CVT           = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_U4),
  OPC_I16U8CVT           = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_U8),
  OPC_I16F4CVT           = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_F4),
  OPC_I16F8CVT           = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_F8),
  OPC_I16FQCVT           = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_FQ),
  OPC_I16F10CVT          = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_F10),
  OPC_I16F16CVT          = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_F16),
  OPC_I16A4CVT           = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_A4),
  OPC_I16A8CVT           = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_A8),
  OPC_I16I16CVT          = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_I16),
  OPC_I16U16CVT          = OPR_CVT + RTYPE(MTYPE_I16) + DESC(MTYPE_U16),
  OPC_U16I4CVT           = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_I4),
  OPC_U16I8CVT           = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_I8),
  OPC_U16U4CVT           = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_U4),
  OPC_U16U8CVT           = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_U8),
  OPC_U16F4CVT           = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_F4),
  OPC_U16F8CVT           = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_F8),
  OPC_U16FQCVT           = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_FQ),
  OPC_U16F10CVT          = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_F10),
  OPC_U16F16CVT          = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_F16),
  OPC_U16A4CVT           = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_A4),
  OPC_U16A8CVT           = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_A8),
  OPC_U16I16CVT          = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_I16),
  OPC_U16U16CVT          = OPR_CVT + RTYPE(MTYPE_U16) + DESC(MTYPE_U16),
  OPC_I4BCVT             = OPR_CVT + RTYPE(MTYPE_I4) + DESC(MTYPE_B),
  OPC_U4BCVT             = OPR_CVT + RTYPE(MTYPE_U4) + DESC(MTYPE_B),
  OPC_I8BCVT             = OPR_CVT + RTYPE(MTYPE_I8) + DESC(MTYPE_B),
  OPC_U8BCVT             = OPR_CVT + RTYPE(MTYPE_U8) + DESC(MTYPE_B),
  OPC_I4CVTL             = OPR_CVTL + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8CVTL             = OPR_CVTL + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4CVTL             = OPR_CVTL + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8CVTL             = OPR_CVTL + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16CVTL            = OPR_CVTL + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16CVTL            = OPR_CVTL + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4DIV              = OPR_DIV + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8DIV              = OPR_DIV + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4DIV              = OPR_DIV + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8DIV              = OPR_DIV + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4DIV              = OPR_DIV + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8DIV              = OPR_DIV + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQDIV              = OPR_DIV + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10DIV             = OPR_DIV + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16DIV             = OPR_DIV + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4DIV              = OPR_DIV + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8DIV              = OPR_DIV + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQDIV              = OPR_DIV + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_C10DIV             = OPR_DIV + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16DIV             = OPR_DIV + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16DIV             = OPR_DIV + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16DIV             = OPR_DIV + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4DIVREM           = OPR_DIVREM + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8DIVREM           = OPR_DIVREM + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4DIVREM           = OPR_DIVREM + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8DIVREM           = OPR_DIVREM + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16DIVREM          = OPR_DIVREM + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16DIVREM          = OPR_DIVREM + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_DO_LOOP            = OPR_DO_LOOP + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_DO_WHILE           = OPR_DO_WHILE + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_BI4EQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_I4),
  OPC_BI8EQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_I8),
  OPC_BU4EQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_U4),
  OPC_BU8EQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_U8),
  OPC_BF4EQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_F4),
  OPC_BF8EQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_F8),
  OPC_BFQEQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_FQ),
  OPC_BF10EQ             = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_F10),
  OPC_BF16EQ             = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_F16),
  OPC_BC4EQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_C4),
  OPC_BC8EQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_C8),
  OPC_BCQEQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_CQ),
  OPC_BA4EQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_A4),
  OPC_BA8EQ              = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_A8),
  OPC_BC10EQ             = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_C10),
  OPC_BC16EQ             = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_C16),
  OPC_BI16EQ             = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_I16),
  OPC_BU16EQ             = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_U16),
  OPC_BBEQ		 = OPR_EQ + RTYPE(MTYPE_B) + DESC(MTYPE_B),
  OPC_I4I4EQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4I8EQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_I8),
  OPC_I4U4EQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_U4),
  OPC_I4U8EQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_U8),
  OPC_I4F4EQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_F4),
  OPC_I4F8EQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_F8),
  OPC_I4FQEQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_FQ),
  OPC_I4F10EQ            = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_F10),
  OPC_I4F16EQ            = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_F16),
  OPC_I4C4EQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_C4),
  OPC_I4C8EQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_C8),
  OPC_I4CQEQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_CQ),
  OPC_I4A4EQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_A4),
  OPC_I4A8EQ             = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_A8),
  OPC_I4C10EQ            = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_C10),
  OPC_I4C16EQ            = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_C16),
  OPC_I4I16EQ            = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_I16),
  OPC_I4U16EQ            = OPR_EQ + RTYPE(MTYPE_I4) + DESC(MTYPE_U16),
  OPC_I8I4EQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_I4),
  OPC_I8I8EQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8U4EQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_U4),
  OPC_I8U8EQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_U8),
  OPC_I8F4EQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_F4),
  OPC_I8F8EQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_F8),
  OPC_I8FQEQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_FQ),
  OPC_I8F10EQ            = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_F10),
  OPC_I8F16EQ            = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_F16),
  OPC_I8C4EQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_C4),
  OPC_I8C8EQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_C8),
  OPC_I8CQEQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_CQ),
  OPC_I8A4EQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_A4),
  OPC_I8A8EQ             = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_A8),
  OPC_I8C10EQ            = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_C10),
  OPC_I8C16EQ            = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_C16),
  OPC_I8I16EQ            = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_I16),
  OPC_I8U16EQ            = OPR_EQ + RTYPE(MTYPE_I8) + DESC(MTYPE_U16),
  OPC_U4I4EQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_I4),
  OPC_U4I8EQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_I8),
  OPC_U4U4EQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4U8EQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_U8),
  OPC_U4F4EQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_F4),
  OPC_U4F8EQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_F8),
  OPC_U4FQEQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_FQ),
  OPC_U4F10EQ            = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_F10),
  OPC_U4F16EQ            = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_F16),
  OPC_U4C4EQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_C4),
  OPC_U4C8EQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_C8),
  OPC_U4CQEQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_CQ),
  OPC_U4A4EQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_A4),
  OPC_U4A8EQ             = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_A8),
  OPC_U4C10EQ            = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_C10),
  OPC_U4C16EQ            = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_C16),
  OPC_U4I16EQ            = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_I16),
  OPC_U4U16EQ            = OPR_EQ + RTYPE(MTYPE_U4) + DESC(MTYPE_U16),
  OPC_U8I4EQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_I4),
  OPC_U8I8EQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_I8),
  OPC_U8U4EQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_U4),
  OPC_U8U8EQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8F4EQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_F4),
  OPC_U8F8EQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_F8),
  OPC_U8FQEQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_FQ),
  OPC_U8F10EQ            = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_F10),
  OPC_U8F16EQ            = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_F16),
  OPC_U8C4EQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_C4),
  OPC_U8C8EQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_C8),
  OPC_U8CQEQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_CQ),
  OPC_U8A4EQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_A4),
  OPC_U8A8EQ             = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_A8),
  OPC_U8C10EQ            = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_C10),
  OPC_U8C16EQ            = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_C16),
  OPC_U8I16EQ            = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_I16),
  OPC_U8U16EQ            = OPR_EQ + RTYPE(MTYPE_U8) + DESC(MTYPE_U16),
  OPC_EVAL               = OPR_EVAL + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_EXC_SCOPE_BEGIN    = OPR_EXC_SCOPE_BEGIN + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_EXC_SCOPE_END      = OPR_EXC_SCOPE_END + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_FALSEBR            = OPR_FALSEBR + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4F4FLOOR          = OPR_FLOOR + RTYPE(MTYPE_I4) + DESC(MTYPE_F4),
  OPC_I4F8FLOOR          = OPR_FLOOR + RTYPE(MTYPE_I4) + DESC(MTYPE_F8),
  OPC_I4FQFLOOR          = OPR_FLOOR + RTYPE(MTYPE_I4) + DESC(MTYPE_FQ),
  OPC_I4F10FLOOR         = OPR_FLOOR + RTYPE(MTYPE_I4) + DESC(MTYPE_F10),
  OPC_I4F16FLOOR         = OPR_FLOOR + RTYPE(MTYPE_I4) + DESC(MTYPE_F16),
  OPC_I8F4FLOOR          = OPR_FLOOR + RTYPE(MTYPE_I8) + DESC(MTYPE_F4),
  OPC_I8F8FLOOR          = OPR_FLOOR + RTYPE(MTYPE_I8) + DESC(MTYPE_F8),
  OPC_I8FQFLOOR          = OPR_FLOOR + RTYPE(MTYPE_I8) + DESC(MTYPE_FQ),
  OPC_I8F10FLOOR         = OPR_FLOOR + RTYPE(MTYPE_I8) + DESC(MTYPE_F10),
  OPC_I8F16FLOOR         = OPR_FLOOR + RTYPE(MTYPE_I8) + DESC(MTYPE_F16),
  OPC_U4F4FLOOR          = OPR_FLOOR + RTYPE(MTYPE_U4) + DESC(MTYPE_F4),
  OPC_U4F8FLOOR          = OPR_FLOOR + RTYPE(MTYPE_U4) + DESC(MTYPE_F8),
  OPC_U4FQFLOOR          = OPR_FLOOR + RTYPE(MTYPE_U4) + DESC(MTYPE_FQ),
  OPC_U4F10FLOOR         = OPR_FLOOR + RTYPE(MTYPE_U4) + DESC(MTYPE_F10),
  OPC_U4F16FLOOR         = OPR_FLOOR + RTYPE(MTYPE_U4) + DESC(MTYPE_F16),
  OPC_U8F4FLOOR          = OPR_FLOOR + RTYPE(MTYPE_U8) + DESC(MTYPE_F4),
  OPC_U8F8FLOOR          = OPR_FLOOR + RTYPE(MTYPE_U8) + DESC(MTYPE_F8),
  OPC_U8FQFLOOR          = OPR_FLOOR + RTYPE(MTYPE_U8) + DESC(MTYPE_FQ),
  OPC_U8F10FLOOR         = OPR_FLOOR + RTYPE(MTYPE_U8) + DESC(MTYPE_F10),
  OPC_U8F16FLOOR         = OPR_FLOOR + RTYPE(MTYPE_U8) + DESC(MTYPE_F16),
  OPC_I16F4FLOOR         = OPR_FLOOR + RTYPE(MTYPE_I16) + DESC(MTYPE_F4),
  OPC_I16F8FLOOR         = OPR_FLOOR + RTYPE(MTYPE_I16) + DESC(MTYPE_F8),
  OPC_I16FQFLOOR         = OPR_FLOOR + RTYPE(MTYPE_I16) + DESC(MTYPE_FQ),
  OPC_I16F10FLOOR        = OPR_FLOOR + RTYPE(MTYPE_I16) + DESC(MTYPE_F10),
  OPC_I16F16FLOOR        = OPR_FLOOR + RTYPE(MTYPE_I16) + DESC(MTYPE_F16),
  OPC_U16F4FLOOR         = OPR_FLOOR + RTYPE(MTYPE_U16) + DESC(MTYPE_F4),
  OPC_U16F8FLOOR         = OPR_FLOOR + RTYPE(MTYPE_U16) + DESC(MTYPE_F8),
  OPC_U16FQFLOOR         = OPR_FLOOR + RTYPE(MTYPE_U16) + DESC(MTYPE_FQ),
  OPC_U16F10FLOOR        = OPR_FLOOR + RTYPE(MTYPE_U16) + DESC(MTYPE_F10),
  OPC_U16F16FLOOR        = OPR_FLOOR + RTYPE(MTYPE_U16) + DESC(MTYPE_F16),
  OPC_F4F4FLOOR          = OPR_FLOOR + RTYPE(MTYPE_F8) + DESC(MTYPE_F4),
  OPC_F8F8FLOOR          = OPR_FLOOR + RTYPE(MTYPE_F8) + DESC(MTYPE_F8),
  OPC_FORWARD_BARRIER    = OPR_FORWARD_BARRIER + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_FUNC_ENTRY         = OPR_FUNC_ENTRY + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_BI4GE              = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_I4),
  OPC_BI8GE              = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_I8),
  OPC_BU4GE              = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_U4),
  OPC_BU8GE              = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_U8),
  OPC_BF4GE              = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_F4),
  OPC_BF8GE              = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_F8),
  OPC_BFQGE              = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_FQ),
  OPC_BF10GE             = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_F10),
  OPC_BF16GE             = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_F16),
  OPC_BA4GE              = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_A4),
  OPC_BA8GE              = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_A8),
  OPC_BI16GE             = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_I16),
  OPC_BU16GE             = OPR_GE + RTYPE(MTYPE_B) + DESC(MTYPE_U16),
  OPC_I4I4GE             = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4I8GE             = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_I8),
  OPC_I4U4GE             = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_U4),
  OPC_I4U8GE             = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_U8),
  OPC_I4F4GE             = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_F4),
  OPC_I4F8GE             = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_F8),
  OPC_I4FQGE             = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_FQ),
  OPC_I4F10GE            = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_F10),
  OPC_I4F16GE            = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_F16),
  OPC_I4A4GE             = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_A4),
  OPC_I4A8GE             = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_A8),
  OPC_I4I16GE            = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_I16),
  OPC_I4U16GE            = OPR_GE + RTYPE(MTYPE_I4) + DESC(MTYPE_U16),
  OPC_I8I4GE             = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_I4),
  OPC_I8I8GE             = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8U4GE             = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_U4),
  OPC_I8U8GE             = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_U8),
  OPC_I8F4GE             = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_F4),
  OPC_I8F8GE             = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_F8),
  OPC_I8FQGE             = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_FQ),
  OPC_I8F10GE            = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_F10),
  OPC_I8F16GE            = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_F16),
  OPC_I8A4GE             = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_A4),
  OPC_I8A8GE             = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_A8),
  OPC_I8I16GE            = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_I16),
  OPC_I8U16GE            = OPR_GE + RTYPE(MTYPE_I8) + DESC(MTYPE_U16),
  OPC_U4I4GE             = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_I4),
  OPC_U4I8GE             = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_I8),
  OPC_U4U4GE             = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4U8GE             = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_U8),
  OPC_U4F4GE             = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_F4),
  OPC_U4F8GE             = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_F8),
  OPC_U4FQGE             = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_FQ),
  OPC_U4F10GE            = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_F10),
  OPC_U4F16GE            = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_F16),
  OPC_U4A4GE             = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_A4),
  OPC_U4A8GE             = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_A8),
  OPC_U4I16GE            = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_I16),
  OPC_U4U16GE            = OPR_GE + RTYPE(MTYPE_U4) + DESC(MTYPE_U16),
  OPC_U8I4GE             = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_I4),
  OPC_U8I8GE             = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_I8),
  OPC_U8U4GE             = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_U4),
  OPC_U8U8GE             = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8F4GE             = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_F4),
  OPC_U8F8GE             = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_F8),
  OPC_U8FQGE             = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_FQ),
  OPC_U8F10GE            = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_F10),
  OPC_U8F16GE            = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_F16),
  OPC_U8A4GE             = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_A4),
  OPC_U8A8GE             = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_A8),
  OPC_U8I16GE            = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_I16),
  OPC_U8U16GE            = OPR_GE + RTYPE(MTYPE_U8) + DESC(MTYPE_U16),
  OPC_GOTO               = OPR_GOTO + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_BI4GT              = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_I4),
  OPC_BI8GT              = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_I8),
  OPC_BU4GT              = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_U4),
  OPC_BU8GT              = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_U8),
  OPC_BF4GT              = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_F4),
  OPC_BF8GT              = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_F8),
  OPC_BFQGT              = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_FQ),
  OPC_BF10GT             = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_F10),
  OPC_BF16GT             = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_F16),
  OPC_BA4GT              = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_A4),
  OPC_BA8GT              = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_A8),
  OPC_BI16GT             = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_I16),
  OPC_BU16GT             = OPR_GT + RTYPE(MTYPE_B) + DESC(MTYPE_U16),
  OPC_I4I4GT             = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4I8GT             = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_I8),
  OPC_I4U4GT             = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_U4),
  OPC_I4U8GT             = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_U8),
  OPC_I4F4GT             = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_F4),
  OPC_I4F8GT             = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_F8),
  OPC_I4FQGT             = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_FQ),
  OPC_I4F10GT            = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_F10),
  OPC_I4F16GT            = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_F16),
  OPC_I4A4GT             = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_A4),
  OPC_I4A8GT             = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_A8),
  OPC_I4I16GT            = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_I16),
  OPC_I4U16GT            = OPR_GT + RTYPE(MTYPE_I4) + DESC(MTYPE_U16),
  OPC_I8I4GT             = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_I4),
  OPC_I8I8GT             = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8U4GT             = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_U4),
  OPC_I8U8GT             = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_U8),
  OPC_I8F4GT             = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_F4),
  OPC_I8F8GT             = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_F8),
  OPC_I8FQGT             = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_FQ),
  OPC_I8F10GT            = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_F10),
  OPC_I8F16GT            = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_F16),
  OPC_I8A4GT             = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_A4),
  OPC_I8A8GT             = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_A8),
  OPC_I8I16GT            = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_I16),
  OPC_I8U16GT            = OPR_GT + RTYPE(MTYPE_I8) + DESC(MTYPE_U16),
  OPC_U4I4GT             = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_I4),
  OPC_U4I8GT             = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_I8),
  OPC_U4U4GT             = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4U8GT             = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_U8),
  OPC_U4F4GT             = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_F4),
  OPC_U4F8GT             = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_F8),
  OPC_U4FQGT             = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_FQ),
  OPC_U4F10GT            = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_F10),
  OPC_U4F16GT            = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_F16),
  OPC_U4A4GT             = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_A4),
  OPC_U4A8GT             = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_A8),
  OPC_U4I16GT            = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_I16),
  OPC_U4U16GT            = OPR_GT + RTYPE(MTYPE_U4) + DESC(MTYPE_U16),
  OPC_U8I4GT             = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_I4),
  OPC_U8I8GT             = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_I8),
  OPC_U8U4GT             = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_U4),
  OPC_U8U8GT             = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8F4GT             = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_F4),
  OPC_U8F8GT             = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_F8),
  OPC_U8FQGT             = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_FQ),
  OPC_U8F10GT            = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_F10),
  OPC_U8F16GT            = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_F16),
  OPC_U8A4GT             = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_A4),
  OPC_U8A8GT             = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_A8),
  OPC_U8I16GT            = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_I16),
  OPC_U8U16GT            = OPR_GT + RTYPE(MTYPE_U8) + DESC(MTYPE_U16),
  OPC_I4HIGHMPY          = OPR_HIGHMPY + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8HIGHMPY          = OPR_HIGHMPY + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4HIGHMPY          = OPR_HIGHMPY + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8HIGHMPY          = OPR_HIGHMPY + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16HIGHMPY         = OPR_HIGHMPY + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16HIGHMPY         = OPR_HIGHMPY + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4HIGHPART         = OPR_HIGHPART + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8HIGHPART         = OPR_HIGHPART + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4HIGHPART         = OPR_HIGHPART + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8HIGHPART         = OPR_HIGHPART + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16HIGHPART        = OPR_HIGHPART + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16HIGHPART        = OPR_HIGHPART + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I1ICALL            = OPR_ICALL + RTYPE(MTYPE_I1) + DESC(MTYPE_V),
  OPC_I2ICALL            = OPR_ICALL + RTYPE(MTYPE_I2) + DESC(MTYPE_V),
  OPC_I4I4ICALL          = OPR_ICALL + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4ICALL            = OPR_ICALL + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8I8ICALL          = OPR_ICALL + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8ICALL            = OPR_ICALL + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U1ICALL            = OPR_ICALL + RTYPE(MTYPE_U1) + DESC(MTYPE_V),
  OPC_U2ICALL            = OPR_ICALL + RTYPE(MTYPE_U2) + DESC(MTYPE_V),
  OPC_U4U4ICALL          = OPR_ICALL + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4ICALL            = OPR_ICALL + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8U8ICALL          = OPR_ICALL + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8ICALL            = OPR_ICALL + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4F4ICALL          = OPR_ICALL + RTYPE(MTYPE_F4) + DESC(MTYPE_F4),
  OPC_F4F8ICALL          = OPR_ICALL + RTYPE(MTYPE_F4) + DESC(MTYPE_F8),
  OPC_F4ICALL            = OPR_ICALL + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8F4ICALL          = OPR_ICALL + RTYPE(MTYPE_F8) + DESC(MTYPE_F4),
  OPC_F8F8ICALL          = OPR_ICALL + RTYPE(MTYPE_F8) + DESC(MTYPE_F8),
  OPC_F8ICALL            = OPR_ICALL + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQICALL            = OPR_ICALL + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10ICALL           = OPR_ICALL + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16ICALL           = OPR_ICALL + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4ICALL            = OPR_ICALL + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8ICALL            = OPR_ICALL + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQICALL            = OPR_ICALL + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MICALL             = OPR_ICALL + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_VICALL             = OPR_ICALL + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_A4ICALL            = OPR_ICALL + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8ICALL            = OPR_ICALL + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10ICALL           = OPR_ICALL + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16ICALL           = OPR_ICALL + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_IDNAME             = OPR_IDNAME + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_IF                 = OPR_IF + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_U4ILDA             = OPR_ILDA + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8ILDA             = OPR_ILDA + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_A4ILDA             = OPR_ILDA + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8ILDA             = OPR_ILDA + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_I4I1ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_I4) + DESC(MTYPE_I1),
  OPC_I4I2ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_I4) + DESC(MTYPE_I2),
  OPC_I4U1ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_I4) + DESC(MTYPE_U1),
  OPC_I4U2ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_I4) + DESC(MTYPE_U2),
  OPC_I8I1ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_I8) + DESC(MTYPE_I1),
  OPC_I8I2ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_I8) + DESC(MTYPE_I2),
  OPC_I8U1ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_I8) + DESC(MTYPE_U1),
  OPC_I8U2ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_I8) + DESC(MTYPE_U2),
  OPC_U4I1ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_U4) + DESC(MTYPE_I1),
  OPC_U4I2ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_U4) + DESC(MTYPE_I2),
  OPC_U4U1ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_U4) + DESC(MTYPE_U1),
  OPC_U4U2ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_U4) + DESC(MTYPE_U2),
  OPC_U8I1ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_U8) + DESC(MTYPE_I1),
  OPC_U8I2ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_U8) + DESC(MTYPE_I2),
  OPC_U8U1ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_U8) + DESC(MTYPE_U1),
  OPC_U8U2ILDBITS        = OPR_ILDBITS + RTYPE(MTYPE_U8) + DESC(MTYPE_U2),
  OPC_I16I1ILDBITS       = OPR_ILDBITS + RTYPE(MTYPE_I16) + DESC(MTYPE_I1),
  OPC_I16I2ILDBITS       = OPR_ILDBITS + RTYPE(MTYPE_I16) + DESC(MTYPE_I2),
  OPC_I16U1ILDBITS       = OPR_ILDBITS + RTYPE(MTYPE_I16) + DESC(MTYPE_U1),
  OPC_I16U2ILDBITS       = OPR_ILDBITS + RTYPE(MTYPE_I16) + DESC(MTYPE_U2),
  OPC_U16I1ILDBITS       = OPR_ILDBITS + RTYPE(MTYPE_U16) + DESC(MTYPE_I1),
  OPC_U16I2ILDBITS       = OPR_ILDBITS + RTYPE(MTYPE_U16) + DESC(MTYPE_I2),
  OPC_U16U1ILDBITS       = OPR_ILDBITS + RTYPE(MTYPE_U16) + DESC(MTYPE_U1),
  OPC_U16U2ILDBITS       = OPR_ILDBITS + RTYPE(MTYPE_U16) + DESC(MTYPE_U2),
  OPC_I4I1ILOAD          = OPR_ILOAD + RTYPE(MTYPE_I4) + DESC(MTYPE_I1),
  OPC_I4I2ILOAD          = OPR_ILOAD + RTYPE(MTYPE_I4) + DESC(MTYPE_I2),
  OPC_I4I4ILOAD          = OPR_ILOAD + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4I8ILOAD          = OPR_ILOAD + RTYPE(MTYPE_I4) + DESC(MTYPE_I8),
  OPC_I4BSILOAD          = OPR_ILOAD + RTYPE(MTYPE_I4) + DESC(MTYPE_BS),
  OPC_I8I1ILOAD          = OPR_ILOAD + RTYPE(MTYPE_I8) + DESC(MTYPE_I1),
  OPC_I8I2ILOAD          = OPR_ILOAD + RTYPE(MTYPE_I8) + DESC(MTYPE_I2),
  OPC_I8I4ILOAD          = OPR_ILOAD + RTYPE(MTYPE_I8) + DESC(MTYPE_I4),
  OPC_I8I8ILOAD          = OPR_ILOAD + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8BSILOAD          = OPR_ILOAD + RTYPE(MTYPE_I8) + DESC(MTYPE_BS),
  OPC_U4U1ILOAD          = OPR_ILOAD + RTYPE(MTYPE_U4) + DESC(MTYPE_U1),
  OPC_U4U2ILOAD          = OPR_ILOAD + RTYPE(MTYPE_U4) + DESC(MTYPE_U2),
  OPC_U4U4ILOAD          = OPR_ILOAD + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4U8ILOAD          = OPR_ILOAD + RTYPE(MTYPE_U4) + DESC(MTYPE_U8),
  OPC_U4BSILOAD          = OPR_ILOAD + RTYPE(MTYPE_U4) + DESC(MTYPE_BS),
  OPC_U8U1ILOAD          = OPR_ILOAD + RTYPE(MTYPE_U8) + DESC(MTYPE_U1),
  OPC_U8U2ILOAD          = OPR_ILOAD + RTYPE(MTYPE_U8) + DESC(MTYPE_U2),
  OPC_U8U4ILOAD          = OPR_ILOAD + RTYPE(MTYPE_U8) + DESC(MTYPE_U4),
  OPC_U8U8ILOAD          = OPR_ILOAD + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8BSILOAD          = OPR_ILOAD + RTYPE(MTYPE_U8) + DESC(MTYPE_BS),
  OPC_F4F4ILOAD          = OPR_ILOAD + RTYPE(MTYPE_F4) + DESC(MTYPE_F4),
  OPC_F8F8ILOAD          = OPR_ILOAD + RTYPE(MTYPE_F8) + DESC(MTYPE_F8),
  OPC_FQFQILOAD          = OPR_ILOAD + RTYPE(MTYPE_FQ) + DESC(MTYPE_FQ),
  OPC_F10F10ILOAD        = OPR_ILOAD + RTYPE(MTYPE_F10) + DESC(MTYPE_F10),
  OPC_F16F16ILOAD        = OPR_ILOAD + RTYPE(MTYPE_F16) + DESC(MTYPE_F16),
  OPC_C4C4ILOAD          = OPR_ILOAD + RTYPE(MTYPE_C4) + DESC(MTYPE_C4),
  OPC_C8C8ILOAD          = OPR_ILOAD + RTYPE(MTYPE_C8) + DESC(MTYPE_C8),
  OPC_CQCQILOAD          = OPR_ILOAD + RTYPE(MTYPE_CQ) + DESC(MTYPE_CQ),
  OPC_MMILOAD            = OPR_ILOAD + RTYPE(MTYPE_M) + DESC(MTYPE_M),
  OPC_A4A4ILOAD          = OPR_ILOAD + RTYPE(MTYPE_A4) + DESC(MTYPE_A4),
  OPC_A8A8ILOAD          = OPR_ILOAD + RTYPE(MTYPE_A8) + DESC(MTYPE_A8),
  OPC_C10C10ILOAD        = OPR_ILOAD + RTYPE(MTYPE_C10) + DESC(MTYPE_C10),
  OPC_C16C16ILOAD        = OPR_ILOAD + RTYPE(MTYPE_C16) + DESC(MTYPE_C16),
  OPC_F4ILOADX           = OPR_ILOADX + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8ILOADX           = OPR_ILOADX + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQILOADX           = OPR_ILOADX + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10ILOADX          = OPR_ILOADX + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16ILOADX          = OPR_ILOADX + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_F4SECONDPART       = OPR_SECONDPART + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8SECONDPART       = OPR_SECONDPART + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQSECONDPART       = OPR_SECONDPART + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10SECONDPART      = OPR_SECONDPART + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16SECONDPART      = OPR_SECONDPART + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_BINTCONST          = OPR_INTCONST + RTYPE(MTYPE_B) + DESC(MTYPE_V),
  OPC_I4INTCONST         = OPR_INTCONST + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8INTCONST         = OPR_INTCONST + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4INTCONST         = OPR_INTCONST + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8INTCONST         = OPR_INTCONST + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_A4INTCONST         = OPR_INTCONST + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8INTCONST         = OPR_INTCONST + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_I16INTCONST        = OPR_INTCONST + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16INTCONST        = OPR_INTCONST + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I1INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_I1) + DESC(MTYPE_V),
  OPC_I2INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_I2) + DESC(MTYPE_V),
  OPC_I4I4INTRINSIC_CALL = OPR_INTRINSIC_CALL + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8I8INTRINSIC_CALL = OPR_INTRINSIC_CALL + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U1INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_U1) + DESC(MTYPE_V),
  OPC_U2INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_U2) + DESC(MTYPE_V),
  OPC_U4U4INTRINSIC_CALL = OPR_INTRINSIC_CALL + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8U8INTRINSIC_CALL = OPR_INTRINSIC_CALL + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4F4INTRINSIC_CALL = OPR_INTRINSIC_CALL + RTYPE(MTYPE_F4) + DESC(MTYPE_F4),
  OPC_F4F8INTRINSIC_CALL = OPR_INTRINSIC_CALL + RTYPE(MTYPE_F4) + DESC(MTYPE_F8),
  OPC_F4INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8F4INTRINSIC_CALL = OPR_INTRINSIC_CALL + RTYPE(MTYPE_F8) + DESC(MTYPE_F4),
  OPC_F8F8INTRINSIC_CALL = OPR_INTRINSIC_CALL + RTYPE(MTYPE_F8) + DESC(MTYPE_F8),
  OPC_F8INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQINTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10INTRINSIC_CALL  = OPR_INTRINSIC_CALL + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16INTRINSIC_CALL  = OPR_INTRINSIC_CALL + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQINTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MINTRINSIC_CALL    = OPR_INTRINSIC_CALL + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_VINTRINSIC_CALL    = OPR_INTRINSIC_CALL + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_A4INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8INTRINSIC_CALL   = OPR_INTRINSIC_CALL + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10INTRINSIC_CALL  = OPR_INTRINSIC_CALL + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16INTRINSIC_CALL  = OPR_INTRINSIC_CALL + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_BINTRINSIC_OP      = OPR_INTRINSIC_OP + RTYPE(MTYPE_B) + DESC(MTYPE_V),
  OPC_I1INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_I1) + DESC(MTYPE_V),
  OPC_I2INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_I2) + DESC(MTYPE_V),
  OPC_I4INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U1INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_U1) + DESC(MTYPE_V),
  OPC_U2INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_U2) + DESC(MTYPE_V),
  OPC_U4INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQINTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10INTRINSIC_OP    = OPR_INTRINSIC_OP + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16INTRINSIC_OP    = OPR_INTRINSIC_OP + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQINTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MINTRINSIC_OP      = OPR_INTRINSIC_OP + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_A4INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8INTRINSIC_OP     = OPR_INTRINSIC_OP + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10INTRINSIC_OP    = OPR_INTRINSIC_OP + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16INTRINSIC_OP    = OPR_INTRINSIC_OP + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16INTRINSIC_OP    = OPR_INTRINSIC_OP + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16INTRINSIC_OP    = OPR_INTRINSIC_OP + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_IO                 = OPR_IO + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_IO_ITEM            = OPR_IO_ITEM + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I1ISTBITS          = OPR_ISTBITS + RTYPE(MTYPE_V) + DESC(MTYPE_I1),
  OPC_I2ISTBITS          = OPR_ISTBITS + RTYPE(MTYPE_V) + DESC(MTYPE_I2),
  OPC_U1ISTBITS          = OPR_ISTBITS + RTYPE(MTYPE_V) + DESC(MTYPE_U1),
  OPC_U2ISTBITS          = OPR_ISTBITS + RTYPE(MTYPE_V) + DESC(MTYPE_U2),
  OPC_I1ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_I1),
  OPC_I2ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_I2),
  OPC_I4ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_I4),
  OPC_I8ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_I8),
  OPC_U1ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_U1),
  OPC_U2ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_U2),
  OPC_U4ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_U4),
  OPC_U8ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_U8),
  OPC_F4ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_F4),
  OPC_F8ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_F8),
  OPC_FQISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_FQ),
  OPC_F10ISTORE          = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_F10),
  OPC_F16ISTORE          = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_F16),
  OPC_C4ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_C4),
  OPC_C8ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_C8),
  OPC_CQISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_CQ),
  OPC_MISTORE            = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_M),
  OPC_BSISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_BS),
  OPC_A4ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_A4),
  OPC_A8ISTORE           = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_A8),
  OPC_C10ISTORE          = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_C10),
  OPC_C16ISTORE          = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_C16),
  OPC_I16ISTORE          = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_I16),
  OPC_U16ISTORE          = OPR_ISTORE + RTYPE(MTYPE_V) + DESC(MTYPE_U16),


  OPC_F4ISTOREX          = OPR_ISTOREX + RTYPE(MTYPE_V) + DESC(MTYPE_F4),
  OPC_F8ISTOREX          = OPR_ISTOREX + RTYPE(MTYPE_V) + DESC(MTYPE_F8),
  OPC_FQISTOREX          = OPR_ISTOREX + RTYPE(MTYPE_V) + DESC(MTYPE_FQ),
  OPC_F10ISTOREX         = OPR_ISTOREX + RTYPE(MTYPE_V) + DESC(MTYPE_F10),
  OPC_F16ISTOREX         = OPR_ISTOREX + RTYPE(MTYPE_V) + DESC(MTYPE_F16),
  OPC_LABEL              = OPR_LABEL + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_BLAND              = OPR_LAND + RTYPE(MTYPE_B) + DESC(MTYPE_V),
  OPC_I4LAND             = OPR_LAND + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_U4LDA              = OPR_LDA + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8LDA              = OPR_LDA + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_A4LDA              = OPR_LDA + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8LDA              = OPR_LDA + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_I4I1LDBITS         = OPR_LDBITS + RTYPE(MTYPE_I4) + DESC(MTYPE_I1),
  OPC_I4I2LDBITS         = OPR_LDBITS + RTYPE(MTYPE_I4) + DESC(MTYPE_I2),
  OPC_I4U1LDBITS         = OPR_LDBITS + RTYPE(MTYPE_I4) + DESC(MTYPE_U1),
  OPC_I4U2LDBITS         = OPR_LDBITS + RTYPE(MTYPE_I4) + DESC(MTYPE_U2),
  OPC_I8I1LDBITS         = OPR_LDBITS + RTYPE(MTYPE_I8) + DESC(MTYPE_I1),
  OPC_I8I2LDBITS         = OPR_LDBITS + RTYPE(MTYPE_I8) + DESC(MTYPE_I2),
  OPC_I8U1LDBITS         = OPR_LDBITS + RTYPE(MTYPE_I8) + DESC(MTYPE_U1),
  OPC_I8U2LDBITS         = OPR_LDBITS + RTYPE(MTYPE_I8) + DESC(MTYPE_U2),
  OPC_U4I1LDBITS         = OPR_LDBITS + RTYPE(MTYPE_U4) + DESC(MTYPE_I1),
  OPC_U4I2LDBITS         = OPR_LDBITS + RTYPE(MTYPE_U4) + DESC(MTYPE_I2),
  OPC_U4U1LDBITS         = OPR_LDBITS + RTYPE(MTYPE_U4) + DESC(MTYPE_U1),
  OPC_U4U2LDBITS         = OPR_LDBITS + RTYPE(MTYPE_U4) + DESC(MTYPE_U2),
  OPC_U8I1LDBITS         = OPR_LDBITS + RTYPE(MTYPE_U8) + DESC(MTYPE_I1),
  OPC_U8I2LDBITS         = OPR_LDBITS + RTYPE(MTYPE_U8) + DESC(MTYPE_I2),
  OPC_U8U1LDBITS         = OPR_LDBITS + RTYPE(MTYPE_U8) + DESC(MTYPE_U1),
  OPC_U8U2LDBITS         = OPR_LDBITS + RTYPE(MTYPE_U8) + DESC(MTYPE_U2),
  OPC_I16I1LDBITS        = OPR_LDBITS + RTYPE(MTYPE_I16) + DESC(MTYPE_I1),
  OPC_I16I2LDBITS        = OPR_LDBITS + RTYPE(MTYPE_I16) + DESC(MTYPE_I2),
  OPC_I16U1LDBITS        = OPR_LDBITS + RTYPE(MTYPE_I16) + DESC(MTYPE_U1),
  OPC_I16U2LDBITS        = OPR_LDBITS + RTYPE(MTYPE_I16) + DESC(MTYPE_U2),
  OPC_U16I1LDBITS        = OPR_LDBITS + RTYPE(MTYPE_U16) + DESC(MTYPE_I1),
  OPC_U16I2LDBITS        = OPR_LDBITS + RTYPE(MTYPE_U16) + DESC(MTYPE_I2),
  OPC_U16U1LDBITS        = OPR_LDBITS + RTYPE(MTYPE_U16) + DESC(MTYPE_U1),
  OPC_U16U2LDBITS        = OPR_LDBITS + RTYPE(MTYPE_U16) + DESC(MTYPE_U2),
  OPC_I4I1LDID           = OPR_LDID + RTYPE(MTYPE_I4) + DESC(MTYPE_I1),
  OPC_I4I2LDID           = OPR_LDID + RTYPE(MTYPE_I4) + DESC(MTYPE_I2),
  OPC_I4I4LDID           = OPR_LDID + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4I8LDID           = OPR_LDID + RTYPE(MTYPE_I4) + DESC(MTYPE_I8),
  OPC_I4BSLDID           = OPR_LDID + RTYPE(MTYPE_I4) + DESC(MTYPE_BS),
  OPC_I8I1LDID           = OPR_LDID + RTYPE(MTYPE_I8) + DESC(MTYPE_I1),
  OPC_I8I2LDID           = OPR_LDID + RTYPE(MTYPE_I8) + DESC(MTYPE_I2),
  OPC_I8I4LDID           = OPR_LDID + RTYPE(MTYPE_I8) + DESC(MTYPE_I4),
  OPC_I8I8LDID           = OPR_LDID + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8BSLDID           = OPR_LDID + RTYPE(MTYPE_I8) + DESC(MTYPE_BS),
  OPC_U4U1LDID           = OPR_LDID + RTYPE(MTYPE_U4) + DESC(MTYPE_U1),
  OPC_U4U2LDID           = OPR_LDID + RTYPE(MTYPE_U4) + DESC(MTYPE_U2),
  OPC_U4U4LDID           = OPR_LDID + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4U8LDID           = OPR_LDID + RTYPE(MTYPE_U4) + DESC(MTYPE_U8),
  OPC_U4BSLDID           = OPR_LDID + RTYPE(MTYPE_U4) + DESC(MTYPE_BS),
  OPC_U8U1LDID           = OPR_LDID + RTYPE(MTYPE_U8) + DESC(MTYPE_U1),
  OPC_U8U2LDID           = OPR_LDID + RTYPE(MTYPE_U8) + DESC(MTYPE_U2),
  OPC_U8U4LDID           = OPR_LDID + RTYPE(MTYPE_U8) + DESC(MTYPE_U4),
  OPC_U8U8LDID           = OPR_LDID + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8BSLDID           = OPR_LDID + RTYPE(MTYPE_U8) + DESC(MTYPE_BS),
  OPC_F4F4LDID           = OPR_LDID + RTYPE(MTYPE_F4) + DESC(MTYPE_F4),
  OPC_F8F8LDID           = OPR_LDID + RTYPE(MTYPE_F8) + DESC(MTYPE_F8),
  OPC_FQFQLDID           = OPR_LDID + RTYPE(MTYPE_FQ) + DESC(MTYPE_FQ),
  OPC_F10F10LDID         = OPR_LDID + RTYPE(MTYPE_F10) + DESC(MTYPE_F10),
  OPC_F16F16LDID         = OPR_LDID + RTYPE(MTYPE_F16) + DESC(MTYPE_F16),
  OPC_C4C4LDID           = OPR_LDID + RTYPE(MTYPE_C4) + DESC(MTYPE_C4),
  OPC_C8C8LDID           = OPR_LDID + RTYPE(MTYPE_C8) + DESC(MTYPE_C8),
  OPC_CQCQLDID           = OPR_LDID + RTYPE(MTYPE_CQ) + DESC(MTYPE_CQ),
  OPC_MMLDID             = OPR_LDID + RTYPE(MTYPE_M) + DESC(MTYPE_M),
  OPC_A4A4LDID           = OPR_LDID + RTYPE(MTYPE_A4) + DESC(MTYPE_A4),
  OPC_A8A8LDID           = OPR_LDID + RTYPE(MTYPE_A8) + DESC(MTYPE_A8),
  OPC_C10C10LDID         = OPR_LDID + RTYPE(MTYPE_C10) + DESC(MTYPE_C10),
  OPC_C16C16LDID         = OPR_LDID + RTYPE(MTYPE_C16) + DESC(MTYPE_C16),
  OPC_BI4LE              = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_I4),
  OPC_BI8LE              = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_I8),
  OPC_BU4LE              = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_U4),
  OPC_BU8LE              = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_U8),
  OPC_BF4LE              = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_F4),
  OPC_BF8LE              = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_F8),
  OPC_BFQLE              = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_FQ),
  OPC_BF10LE             = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_F10),
  OPC_BF16LE             = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_F16),
  OPC_BA4LE              = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_A4),
  OPC_BA8LE              = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_A8),
  OPC_BI16LE             = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_I16),
  OPC_BU16LE             = OPR_LE + RTYPE(MTYPE_B) + DESC(MTYPE_U16),
  OPC_I4I4LE             = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4I8LE             = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_I8),
  OPC_I4U4LE             = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_U4),
  OPC_I4U8LE             = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_U8),
  OPC_I4F4LE             = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_F4),
  OPC_I4F8LE             = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_F8),
  OPC_I4FQLE             = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_FQ),
  OPC_I4F10LE            = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_F10),
  OPC_I4F16LE            = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_F16),
  OPC_I4A4LE             = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_A4),
  OPC_I4A8LE             = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_A8),
  OPC_I4I16LE            = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_I16),
  OPC_I4U16LE            = OPR_LE + RTYPE(MTYPE_I4) + DESC(MTYPE_U16),
  OPC_I8I4LE             = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_I4),
  OPC_I8I8LE             = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8U4LE             = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_U4),
  OPC_I8U8LE             = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_U8),
  OPC_I8F4LE             = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_F4),
  OPC_I8F8LE             = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_F8),
  OPC_I8FQLE             = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_FQ),
  OPC_I8F10LE            = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_F10),
  OPC_I8F16LE            = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_F16),
  OPC_I8A4LE             = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_A4),
  OPC_I8A8LE             = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_A8),
  OPC_I8I16LE            = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_I16),
  OPC_I8U16LE            = OPR_LE + RTYPE(MTYPE_I8) + DESC(MTYPE_U16),
  OPC_U4I4LE             = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_I4),
  OPC_U4I8LE             = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_I8),
  OPC_U4U4LE             = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4U8LE             = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_U8),
  OPC_U4F4LE             = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_F4),
  OPC_U4F8LE             = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_F8),
  OPC_U4FQLE             = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_FQ),
  OPC_U4F10LE            = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_F10),
  OPC_U4F16LE            = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_F16),
  OPC_U4A4LE             = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_A4),
  OPC_U4A8LE             = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_A8),
  OPC_U4I16LE            = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_I16),
  OPC_U4U16LE            = OPR_LE + RTYPE(MTYPE_U4) + DESC(MTYPE_U16),
  OPC_U8I4LE             = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_I4),
  OPC_U8I8LE             = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_I8),
  OPC_U8U4LE             = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_U4),
  OPC_U8U8LE             = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8F4LE             = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_F4),
  OPC_U8F8LE             = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_F8),
  OPC_U8FQLE             = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_FQ),
  OPC_U8F10LE            = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_F10),
  OPC_U8F16LE            = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_F16),
  OPC_U8A4LE             = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_A4),
  OPC_U8A8LE             = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_A8),
  OPC_U8I16LE            = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_I16),
  OPC_U8U16LE            = OPR_LE + RTYPE(MTYPE_U8) + DESC(MTYPE_U16),
  OPC_BLIOR              = OPR_LIOR + RTYPE(MTYPE_B) + DESC(MTYPE_V),
  OPC_I4LIOR             = OPR_LIOR + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_BLNOT              = OPR_LNOT + RTYPE(MTYPE_B) + DESC(MTYPE_V),
  OPC_I4LNOT             = OPR_LNOT + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_LOOP_INFO          = OPR_LOOP_INFO + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4LOWPART          = OPR_LOWPART + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8LOWPART          = OPR_LOWPART + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4LOWPART          = OPR_LOWPART + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8LOWPART          = OPR_LOWPART + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16LOWPART         = OPR_LOWPART + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16LOWPART         = OPR_LOWPART + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4LSHR             = OPR_LSHR + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8LSHR             = OPR_LSHR + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4LSHR             = OPR_LSHR + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8LSHR             = OPR_LSHR + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16LSHR            = OPR_LSHR + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16LSHR            = OPR_LSHR + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_BI4LT              = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_I4),
  OPC_BI8LT              = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_I8),
  OPC_BU4LT              = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_U4),
  OPC_BU8LT              = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_U8),
  OPC_BF4LT              = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_F4),
  OPC_BF8LT              = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_F8),
  OPC_BFQLT              = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_FQ),
  OPC_BF10LT             = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_F10),
  OPC_BF16LT             = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_F16),
  OPC_BA4LT              = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_A4),
  OPC_BA8LT              = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_A8),
  OPC_BI16LT             = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_I16),
  OPC_BU16LT             = OPR_LT + RTYPE(MTYPE_B) + DESC(MTYPE_U16),
  OPC_I4I4LT             = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4I8LT             = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_I8),
  OPC_I4U4LT             = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_U4),
  OPC_I4U8LT             = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_U8),
  OPC_I4F4LT             = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_F4),
  OPC_I4F8LT             = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_F8),
  OPC_I4FQLT             = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_FQ),
  OPC_I4F10LT            = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_F10),
  OPC_I4F16LT            = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_F16),
  OPC_I4A4LT             = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_A4),
  OPC_I4A8LT             = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_A8),
  OPC_I4I16LT            = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_I16),
  OPC_I4U16LT            = OPR_LT + RTYPE(MTYPE_I4) + DESC(MTYPE_U16),
  OPC_I8I4LT             = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_I4),
  OPC_I8I8LT             = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8U4LT             = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_U4),
  OPC_I8U8LT             = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_U8),
  OPC_I8F4LT             = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_F4),
  OPC_I8F8LT             = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_F8),
  OPC_I8FQLT             = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_FQ),
  OPC_I8F10LT            = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_F10),
  OPC_I8F16LT            = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_F16),
  OPC_I8A4LT             = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_A4),
  OPC_I8A8LT             = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_A8),
  OPC_I8I16LT            = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_I16),
  OPC_I8U16LT            = OPR_LT + RTYPE(MTYPE_I8) + DESC(MTYPE_U16),
  OPC_U4I4LT             = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_I4),
  OPC_U4I8LT             = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_I8),
  OPC_U4U4LT             = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4U8LT             = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_U8),
  OPC_U4F4LT             = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_F4),
  OPC_U4F8LT             = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_F8),
  OPC_U4FQLT             = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_FQ),
  OPC_U4F10LT            = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_F10),
  OPC_U4F16LT            = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_F16),
  OPC_U4A4LT             = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_A4),
  OPC_U4A8LT             = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_A8),
  OPC_U4I16LT            = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_I16),
  OPC_U4U16LT            = OPR_LT + RTYPE(MTYPE_U4) + DESC(MTYPE_U16),
  OPC_U8I4LT             = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_I4),
  OPC_U8I8LT             = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_I8),
  OPC_U8U4LT             = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_U4),
  OPC_U8U8LT             = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8F4LT             = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_F4),
  OPC_U8F8LT             = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_F8),
  OPC_U8FQLT             = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_FQ),
  OPC_U8F10LT            = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_F10),
  OPC_U8F16LT            = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_F16),
  OPC_U8A4LT             = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_A4),
  OPC_U8A8LT             = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_A8),
  OPC_U8I16LT            = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_I16),
  OPC_U8U16LT            = OPR_LT + RTYPE(MTYPE_U8) + DESC(MTYPE_U16),
  OPC_F4MADD             = OPR_MADD + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8MADD             = OPR_MADD + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQMADD             = OPR_MADD + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10MADD            = OPR_MADD + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16MADD            = OPR_MADD + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_I4MAX              = OPR_MAX + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8MAX              = OPR_MAX + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4MAX              = OPR_MAX + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8MAX              = OPR_MAX + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4MAX              = OPR_MAX + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8MAX              = OPR_MAX + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQMAX              = OPR_MAX + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10MAX             = OPR_MAX + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16MAX             = OPR_MAX + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_A4MAX              = OPR_MAX + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8MAX              = OPR_MAX + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_I16MAX             = OPR_MAX + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16MAX             = OPR_MAX + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4MAXPART          = OPR_MAXPART + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8MAXPART          = OPR_MAXPART + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4MAXPART          = OPR_MAXPART + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8MAXPART          = OPR_MAXPART + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4MAXPART          = OPR_MAXPART + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8MAXPART          = OPR_MAXPART + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQMAXPART          = OPR_MAXPART + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10MAXPART         = OPR_MAXPART + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16MAXPART         = OPR_MAXPART + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_I16MAXPART         = OPR_MAXPART + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16MAXPART         = OPR_MAXPART + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4MIN              = OPR_MIN + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8MIN              = OPR_MIN + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4MIN              = OPR_MIN + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8MIN              = OPR_MIN + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4MIN              = OPR_MIN + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8MIN              = OPR_MIN + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQMIN              = OPR_MIN + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10MIN             = OPR_MIN + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16MIN             = OPR_MIN + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_A4MIN              = OPR_MIN + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8MIN              = OPR_MIN + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_I16MIN             = OPR_MIN + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16MIN             = OPR_MIN + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4MINMAX           = OPR_MINMAX + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8MINMAX           = OPR_MINMAX + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4MINMAX           = OPR_MINMAX + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8MINMAX           = OPR_MINMAX + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4MINMAX           = OPR_MINMAX + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8MINMAX           = OPR_MINMAX + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQMINMAX           = OPR_MINMAX + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10MINMAX          = OPR_MINMAX + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16MINMAX          = OPR_MINMAX + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_A4MINMAX           = OPR_MINMAX + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8MINMAX           = OPR_MINMAX + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_I16MINMAX          = OPR_MINMAX + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16MINMAX          = OPR_MINMAX + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4MINPART          = OPR_MINPART + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8MINPART          = OPR_MINPART + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4MINPART          = OPR_MINPART + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8MINPART          = OPR_MINPART + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4MINPART          = OPR_MINPART + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8MINPART          = OPR_MINPART + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQMINPART          = OPR_MINPART + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10MINPART         = OPR_MINPART + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16MINPART         = OPR_MINPART + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_I16MINPART         = OPR_MINPART + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16MINPART         = OPR_MINPART + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_MLOAD              = OPR_MLOAD + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_I4MOD              = OPR_MOD + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8MOD              = OPR_MOD + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4MOD              = OPR_MOD + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8MOD              = OPR_MOD + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16MOD             = OPR_MOD + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16MOD             = OPR_MOD + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4MPY              = OPR_MPY + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8MPY              = OPR_MPY + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4MPY              = OPR_MPY + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8MPY              = OPR_MPY + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4MPY              = OPR_MPY + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8MPY              = OPR_MPY + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQMPY              = OPR_MPY + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10MPY             = OPR_MPY + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16MPY             = OPR_MPY + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4MPY              = OPR_MPY + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8MPY              = OPR_MPY + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQMPY              = OPR_MPY + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_C10MPY             = OPR_MPY + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16MPY             = OPR_MPY + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16MPY             = OPR_MPY + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16MPY             = OPR_MPY + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_MSTORE             = OPR_MSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_F4MSUB             = OPR_MSUB + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8MSUB             = OPR_MSUB + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQMSUB             = OPR_MSUB + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10MSUB            = OPR_MSUB + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16MSUB            = OPR_MSUB + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_BI4NE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_I4),
  OPC_BI8NE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_I8),
  OPC_BU4NE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_U4),
  OPC_BU8NE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_U8),
  OPC_BF4NE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_F4),
  OPC_BF8NE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_F8),
  OPC_BFQNE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_FQ),
  OPC_BF10NE             = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_F10),
  OPC_BF16NE             = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_F16),
  OPC_BC4NE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_C4),
  OPC_BC8NE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_C8),
  OPC_BCQNE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_CQ),
  OPC_BA4NE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_A4),
  OPC_BA8NE              = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_A8),
  OPC_BC10NE             = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_C10),
  OPC_BC16NE             = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_C16),
  OPC_BI16NE             = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_I16),
  OPC_BU16NE             = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_U16),
  OPC_BBNE		 = OPR_NE + RTYPE(MTYPE_B) + DESC(MTYPE_B),
  OPC_I4I4NE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4I8NE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_I8),
  OPC_I4U4NE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_U4),
  OPC_I4U8NE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_U8),
  OPC_I4F4NE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_F4),
  OPC_I4F8NE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_F8),
  OPC_I4FQNE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_FQ),
  OPC_I4F10NE            = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_F10),
  OPC_I4F16NE            = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_F16),
  OPC_I4C4NE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_C4),
  OPC_I4C8NE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_C8),
  OPC_I4CQNE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_CQ),
  OPC_I4A4NE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_A4),
  OPC_I4A8NE             = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_A8),
  OPC_I4C10NE            = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_C10),
  OPC_I4C16NE            = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_C16),
  OPC_I4I16NE            = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_I16),
  OPC_I4U16NE            = OPR_NE + RTYPE(MTYPE_I4) + DESC(MTYPE_U16),
  OPC_I8I4NE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_I4),
  OPC_I8I8NE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8U4NE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_U4),
  OPC_I8U8NE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_U8),
  OPC_I8F4NE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_F4),
  OPC_I8F8NE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_F8),
  OPC_I8FQNE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_FQ),
  OPC_I8F10NE            = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_F10),
  OPC_I8F16NE            = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_F16),
  OPC_I8C4NE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_C4),
  OPC_I8C8NE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_C8),
  OPC_I8CQNE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_CQ),
  OPC_I8A4NE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_A4),
  OPC_I8A8NE             = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_A8),
  OPC_I8C10NE            = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_C10),
  OPC_I8C16NE            = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_C16),
  OPC_I8I16NE            = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_I16),
  OPC_I8U16NE            = OPR_NE + RTYPE(MTYPE_I8) + DESC(MTYPE_U16),
  OPC_U4I4NE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_I4),
  OPC_U4I8NE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_I8),
  OPC_U4U4NE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4U8NE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_U8),
  OPC_U4F4NE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_F4),
  OPC_U4F8NE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_F8),
  OPC_U4FQNE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_FQ),
  OPC_U4F10NE            = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_F10),
  OPC_U4F16NE            = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_F16),
  OPC_U4C4NE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_C4),
  OPC_U4C8NE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_C8),
  OPC_U4CQNE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_CQ),
  OPC_U4A4NE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_A4),
  OPC_U4A8NE             = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_A8),
  OPC_U4C10NE            = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_C10),
  OPC_U4C16NE            = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_C16),
  OPC_U4I16NE            = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_I16),
  OPC_U4U16NE            = OPR_NE + RTYPE(MTYPE_U4) + DESC(MTYPE_U16),
  OPC_U8I4NE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_I4),
  OPC_U8I8NE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_I8),
  OPC_U8U4NE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_U4),
  OPC_U8U8NE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8F4NE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_F4),
  OPC_U8F8NE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_F8),
  OPC_U8FQNE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_FQ),
  OPC_U8F10NE            = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_F10),
  OPC_U8F16NE            = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_F16),
  OPC_U8C4NE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_C4),
  OPC_U8C8NE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_C8),
  OPC_U8CQNE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_CQ),
  OPC_U8A4NE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_A4),
  OPC_U8A8NE             = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_A8),
  OPC_U8C10NE            = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_C10),
  OPC_U8C16NE            = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_C16),
  OPC_U8I16NE            = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_I16),
  OPC_U8U16NE            = OPR_NE + RTYPE(MTYPE_U8) + DESC(MTYPE_U16),
  OPC_I4NEG              = OPR_NEG + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8NEG              = OPR_NEG + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4NEG              = OPR_NEG + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8NEG              = OPR_NEG + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4NEG              = OPR_NEG + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8NEG              = OPR_NEG + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQNEG              = OPR_NEG + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10NEG             = OPR_NEG + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16NEG             = OPR_NEG + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4NEG              = OPR_NEG + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8NEG              = OPR_NEG + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQNEG              = OPR_NEG + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_C10NEG             = OPR_NEG + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16NEG             = OPR_NEG + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16NEG             = OPR_NEG + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16NEG             = OPR_NEG + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_F4NMADD            = OPR_NMADD + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8NMADD            = OPR_NMADD + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQNMADD            = OPR_NMADD + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10NMADD           = OPR_NMADD + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16NMADD           = OPR_NMADD + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_F4NMSUB            = OPR_NMSUB + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8NMSUB            = OPR_NMSUB + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQNMSUB            = OPR_NMSUB + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10NMSUB           = OPR_NMSUB + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16NMSUB           = OPR_NMSUB + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_I4OPTPARM          = OPR_OPTPARM + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8OPTPARM          = OPR_OPTPARM + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4OPTPARM          = OPR_OPTPARM + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8OPTPARM          = OPR_OPTPARM + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4OPTPARM          = OPR_OPTPARM + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8OPTPARM          = OPR_OPTPARM + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQOPTPARM          = OPR_OPTPARM + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10OPTPARM         = OPR_OPTPARM + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16OPTPARM         = OPR_OPTPARM + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4OPTPARM          = OPR_OPTPARM + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8OPTPARM          = OPR_OPTPARM + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQOPTPARM          = OPR_OPTPARM + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_C10OPTPARM         = OPR_OPTPARM + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16OPTPARM         = OPR_OPTPARM + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16OPTPARM         = OPR_OPTPARM + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16OPTPARM         = OPR_OPTPARM + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_OPT_CHI            = OPR_OPT_CHI + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_OPT_RESERVE2       = OPR_OPT_RESERVE2 + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4PAREN            = OPR_PAREN + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8PAREN            = OPR_PAREN + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4PAREN            = OPR_PAREN + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8PAREN            = OPR_PAREN + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4PAREN            = OPR_PAREN + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8PAREN            = OPR_PAREN + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQPAREN            = OPR_PAREN + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10PAREN           = OPR_PAREN + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16PAREN           = OPR_PAREN + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4PAREN            = OPR_PAREN + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8PAREN            = OPR_PAREN + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQPAREN            = OPR_PAREN + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_C10PAREN           = OPR_PAREN + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16PAREN           = OPR_PAREN + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I4PARM             = OPR_PARM + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8PARM             = OPR_PARM + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4PARM             = OPR_PARM + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8PARM             = OPR_PARM + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4PARM             = OPR_PARM + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8PARM             = OPR_PARM + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQPARM             = OPR_PARM + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10PARM            = OPR_PARM + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16PARM            = OPR_PARM + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4PARM             = OPR_PARM + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8PARM             = OPR_PARM + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQPARM             = OPR_PARM + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MPARM              = OPR_PARM + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_VPARM              = OPR_PARM + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_A4PARM             = OPR_PARM + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8PARM             = OPR_PARM + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10PARM            = OPR_PARM + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16PARM            = OPR_PARM + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16PARM            = OPR_PARM + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16PARM            = OPR_PARM + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I1PICCALL          = OPR_PICCALL + RTYPE(MTYPE_I1) + DESC(MTYPE_V),
  OPC_I2PICCALL          = OPR_PICCALL + RTYPE(MTYPE_I2) + DESC(MTYPE_V),
  OPC_I4I4PICCALL        = OPR_PICCALL + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4PICCALL          = OPR_PICCALL + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8I8PICCALL        = OPR_PICCALL + RTYPE(MTYPE_I8) + DESC(MTYPE_I8),
  OPC_I8PICCALL          = OPR_PICCALL + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U1PICCALL          = OPR_PICCALL + RTYPE(MTYPE_U1) + DESC(MTYPE_V),
  OPC_U2PICCALL          = OPR_PICCALL + RTYPE(MTYPE_U2) + DESC(MTYPE_V),
  OPC_U4U4PICCALL        = OPR_PICCALL + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U4PICCALL          = OPR_PICCALL + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8U8PICCALL        = OPR_PICCALL + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U8PICCALL          = OPR_PICCALL + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4F4PICCALL        = OPR_PICCALL + RTYPE(MTYPE_F4) + DESC(MTYPE_F4),
  OPC_F4F8PICCALL        = OPR_PICCALL + RTYPE(MTYPE_F4) + DESC(MTYPE_F8),
  OPC_F4PICCALL          = OPR_PICCALL + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8F4PICCALL        = OPR_PICCALL + RTYPE(MTYPE_F8) + DESC(MTYPE_F4),
  OPC_F8F8PICCALL        = OPR_PICCALL + RTYPE(MTYPE_F8) + DESC(MTYPE_F8),
  OPC_F8PICCALL          = OPR_PICCALL + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQPICCALL          = OPR_PICCALL + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10PICCALL         = OPR_PICCALL + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16PICCALL         = OPR_PICCALL + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4PICCALL          = OPR_PICCALL + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8PICCALL          = OPR_PICCALL + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_VPICCALL           = OPR_PICCALL + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_A4PICCALL          = OPR_PICCALL + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8PICCALL          = OPR_PICCALL + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_PRAGMA             = OPR_PRAGMA + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_PREFETCH           = OPR_PREFETCH + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_PREFETCHX          = OPR_PREFETCHX + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4RCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8RCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4RCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8RCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4RCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8RCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQRCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10RCOMMA          = OPR_RCOMMA + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16RCOMMA          = OPR_RCOMMA + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4RCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8RCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQRCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MRCOMMA            = OPR_RCOMMA + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_A4RCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8RCOMMA           = OPR_RCOMMA + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10RCOMMA          = OPR_RCOMMA + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16RCOMMA          = OPR_RCOMMA + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16RCOMMA          = OPR_RCOMMA + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16RCOMMA          = OPR_RCOMMA + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_F4FIRSTPART        = OPR_FIRSTPART + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8FIRSTPART        = OPR_FIRSTPART + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQFIRSTPART        = OPR_FIRSTPART + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10FIRSTPART       = OPR_FIRSTPART + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16FIRSTPART       = OPR_FIRSTPART + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_F4RECIP            = OPR_RECIP + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8RECIP            = OPR_RECIP + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQRECIP            = OPR_RECIP + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10RECIP           = OPR_RECIP + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16RECIP           = OPR_RECIP + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4RECIP            = OPR_RECIP + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8RECIP            = OPR_RECIP + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQRECIP            = OPR_RECIP + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_C10RECIP           = OPR_RECIP + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16RECIP           = OPR_RECIP + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_REGION             = OPR_REGION + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_REGION_EXIT        = OPR_REGION_EXIT + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4REM              = OPR_REM + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8REM              = OPR_REM + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4REM              = OPR_REM + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8REM              = OPR_REM + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16REM             = OPR_REM + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16REM             = OPR_REM + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_RETURN             = OPR_RETURN + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I1RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_I1) + DESC(MTYPE_V),
  OPC_I2RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_I2) + DESC(MTYPE_V),
  OPC_I4RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U1RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_U1) + DESC(MTYPE_V),
  OPC_U2RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_U2) + DESC(MTYPE_V),
  OPC_U4RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQRETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10RETURN_VAL      = OPR_RETURN_VAL + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16RETURN_VAL      = OPR_RETURN_VAL + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQRETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MRETURN_VAL        = OPR_RETURN_VAL + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_A4RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8RETURN_VAL       = OPR_RETURN_VAL + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10RETURN_VAL      = OPR_RETURN_VAL + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16RETURN_VAL      = OPR_RETURN_VAL + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16RETURN_VAL      = OPR_RETURN_VAL + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16RETURN_VAL      = OPR_RETURN_VAL + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_I4F4RND            = OPR_RND + RTYPE(MTYPE_I4) + DESC(MTYPE_F4),
  OPC_I4F8RND            = OPR_RND + RTYPE(MTYPE_I4) + DESC(MTYPE_F8),
  OPC_I4FQRND            = OPR_RND + RTYPE(MTYPE_I4) + DESC(MTYPE_FQ),
  OPC_I4F10RND           = OPR_RND + RTYPE(MTYPE_I4) + DESC(MTYPE_F10),
  OPC_I4F16RND           = OPR_RND + RTYPE(MTYPE_I4) + DESC(MTYPE_F16),
  OPC_I8F4RND            = OPR_RND + RTYPE(MTYPE_I8) + DESC(MTYPE_F4),
  OPC_I8F8RND            = OPR_RND + RTYPE(MTYPE_I8) + DESC(MTYPE_F8),
  OPC_I8FQRND            = OPR_RND + RTYPE(MTYPE_I8) + DESC(MTYPE_FQ),
  OPC_I8F10RND           = OPR_RND + RTYPE(MTYPE_I8) + DESC(MTYPE_F10),
  OPC_I8F16RND           = OPR_RND + RTYPE(MTYPE_I8) + DESC(MTYPE_F16),
  OPC_U4F4RND            = OPR_RND + RTYPE(MTYPE_U4) + DESC(MTYPE_F4),
  OPC_U4F8RND            = OPR_RND + RTYPE(MTYPE_U4) + DESC(MTYPE_F8),
  OPC_U4FQRND            = OPR_RND + RTYPE(MTYPE_U4) + DESC(MTYPE_FQ),
  OPC_U4F10RND           = OPR_RND + RTYPE(MTYPE_U4) + DESC(MTYPE_F10),
  OPC_U4F16RND           = OPR_RND + RTYPE(MTYPE_U4) + DESC(MTYPE_F16),
  OPC_U8F4RND            = OPR_RND + RTYPE(MTYPE_U8) + DESC(MTYPE_F4),
  OPC_U8F8RND            = OPR_RND + RTYPE(MTYPE_U8) + DESC(MTYPE_F8),
  OPC_U8FQRND            = OPR_RND + RTYPE(MTYPE_U8) + DESC(MTYPE_FQ),
  OPC_U8F10RND           = OPR_RND + RTYPE(MTYPE_U8) + DESC(MTYPE_F10),
  OPC_U8F16RND           = OPR_RND + RTYPE(MTYPE_U8) + DESC(MTYPE_F16),
  OPC_I16F4RND           = OPR_RND + RTYPE(MTYPE_I16) + DESC(MTYPE_F4),
  OPC_I16F8RND           = OPR_RND + RTYPE(MTYPE_I16) + DESC(MTYPE_F8),
  OPC_I16FQRND           = OPR_RND + RTYPE(MTYPE_I16) + DESC(MTYPE_FQ),
  OPC_I16F10RND          = OPR_RND + RTYPE(MTYPE_I16) + DESC(MTYPE_F10),
  OPC_I16F16RND          = OPR_RND + RTYPE(MTYPE_I16) + DESC(MTYPE_F16),
  OPC_U16F4RND           = OPR_RND + RTYPE(MTYPE_U16) + DESC(MTYPE_F4),
  OPC_U16F8RND           = OPR_RND + RTYPE(MTYPE_U16) + DESC(MTYPE_F8),
  OPC_U16FQRND           = OPR_RND + RTYPE(MTYPE_U16) + DESC(MTYPE_FQ),
  OPC_U16F10RND          = OPR_RND + RTYPE(MTYPE_U16) + DESC(MTYPE_F10),
  OPC_U16F16RND          = OPR_RND + RTYPE(MTYPE_U16) + DESC(MTYPE_F16),
  OPC_F4RSQRT            = OPR_RSQRT + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8RSQRT            = OPR_RSQRT + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQRSQRT            = OPR_RSQRT + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10RSQRT           = OPR_RSQRT + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16RSQRT           = OPR_RSQRT + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4RSQRT            = OPR_RSQRT + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8RSQRT            = OPR_RSQRT + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQRSQRT            = OPR_RSQRT + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_C10RSQRT           = OPR_RSQRT + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16RSQRT           = OPR_RSQRT + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_BSELECT            = OPR_SELECT + RTYPE(MTYPE_B) + DESC(MTYPE_V),
  OPC_I4SELECT           = OPR_SELECT + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8SELECT           = OPR_SELECT + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4SELECT           = OPR_SELECT + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8SELECT           = OPR_SELECT + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4SELECT           = OPR_SELECT + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8SELECT           = OPR_SELECT + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQSELECT           = OPR_SELECT + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10SELECT          = OPR_SELECT + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16SELECT          = OPR_SELECT + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4SELECT           = OPR_SELECT + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8SELECT           = OPR_SELECT + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQSELECT           = OPR_SELECT + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_A4SELECT           = OPR_SELECT + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8SELECT           = OPR_SELECT + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10SELECT          = OPR_SELECT + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16SELECT          = OPR_SELECT + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16SELECT          = OPR_SELECT + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16SELECT          = OPR_SELECT + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_BBSELECT           = OPR_SELECT + RTYPE(MTYPE_B) + DESC(MTYPE_B),
  OPC_I4BSELECT          = OPR_SELECT + RTYPE(MTYPE_I4) + DESC(MTYPE_B),
  OPC_I8BSELECT          = OPR_SELECT + RTYPE(MTYPE_I8) + DESC(MTYPE_B),
  OPC_U4BSELECT          = OPR_SELECT + RTYPE(MTYPE_U4) + DESC(MTYPE_B),
  OPC_U8BSELECT          = OPR_SELECT + RTYPE(MTYPE_U8) + DESC(MTYPE_B),
  OPC_F4BSELECT          = OPR_SELECT + RTYPE(MTYPE_F4) + DESC(MTYPE_B),
  OPC_F8BSELECT          = OPR_SELECT + RTYPE(MTYPE_F8) + DESC(MTYPE_B),
  OPC_FQBSELECT          = OPR_SELECT + RTYPE(MTYPE_FQ) + DESC(MTYPE_B),
  OPC_F10BSELECT         = OPR_SELECT + RTYPE(MTYPE_F10) + DESC(MTYPE_B),
  OPC_F16BSELECT         = OPR_SELECT + RTYPE(MTYPE_F16) + DESC(MTYPE_B),
  OPC_C4BSELECT          = OPR_SELECT + RTYPE(MTYPE_C4) + DESC(MTYPE_B),
  OPC_C8BSELECT          = OPR_SELECT + RTYPE(MTYPE_C8) + DESC(MTYPE_B),
  OPC_CQBSELECT          = OPR_SELECT + RTYPE(MTYPE_CQ) + DESC(MTYPE_B),
  OPC_A4BSELECT          = OPR_SELECT + RTYPE(MTYPE_A4) + DESC(MTYPE_B),
  OPC_A8BSELECT          = OPR_SELECT + RTYPE(MTYPE_A8) + DESC(MTYPE_B),
  OPC_C10BSELECT         = OPR_SELECT + RTYPE(MTYPE_C10) + DESC(MTYPE_B),
  OPC_C16BSELECT         = OPR_SELECT + RTYPE(MTYPE_C16) + DESC(MTYPE_B),
  OPC_I16BSELECT         = OPR_SELECT + RTYPE(MTYPE_I16) + DESC(MTYPE_B),
  OPC_U16BSELECT         = OPR_SELECT + RTYPE(MTYPE_U16) + DESC(MTYPE_B),
  OPC_I4SHL              = OPR_SHL + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8SHL              = OPR_SHL + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4SHL              = OPR_SHL + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8SHL              = OPR_SHL + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16SHL             = OPR_SHL + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16SHL             = OPR_SHL + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_F4SQRT             = OPR_SQRT + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8SQRT             = OPR_SQRT + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQSQRT             = OPR_SQRT + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10SQRT            = OPR_SQRT + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16SQRT            = OPR_SQRT + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4SQRT             = OPR_SQRT + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8SQRT             = OPR_SQRT + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQSQRT             = OPR_SQRT + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_C10SQRT            = OPR_SQRT + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16SQRT            = OPR_SQRT + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I1STBITS           = OPR_STBITS + RTYPE(MTYPE_V) + DESC(MTYPE_I1),
  OPC_I2STBITS           = OPR_STBITS + RTYPE(MTYPE_V) + DESC(MTYPE_I2),
  OPC_U1STBITS           = OPR_STBITS + RTYPE(MTYPE_V) + DESC(MTYPE_U1),
  OPC_U2STBITS           = OPR_STBITS + RTYPE(MTYPE_V) + DESC(MTYPE_U2),
  OPC_I1STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_I1),
  OPC_I2STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_I2),
  OPC_I4STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_I4),
  OPC_I8STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_I8),
  OPC_U1STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_U1),
  OPC_U2STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_U2),
  OPC_U4STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_U4),
  OPC_U8STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_U8),
  OPC_F4STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_F4),
  OPC_F8STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_F8),
  OPC_FQSTID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_FQ),
  OPC_F10STID            = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_F10),
  OPC_F16STID            = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_F16),
  OPC_C4STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_C4),
  OPC_C8STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_C8),
  OPC_CQSTID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_CQ),
  OPC_MSTID              = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_M),
  OPC_BSSTID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_BS),
  OPC_A4STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_A4),
  OPC_A8STID             = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_A8),
  OPC_C10STID            = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_C10),
  OPC_C16STID            = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_C16),
  OPC_I16STID            = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_I16),
  OPC_U16STID            = OPR_STID + RTYPE(MTYPE_V) + DESC(MTYPE_U16),
  OPC_I4SUB              = OPR_SUB + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8SUB              = OPR_SUB + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4SUB              = OPR_SUB + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8SUB              = OPR_SUB + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4SUB              = OPR_SUB + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8SUB              = OPR_SUB + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQSUB              = OPR_SUB + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10SUB             = OPR_SUB + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16SUB             = OPR_SUB + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4SUB              = OPR_SUB + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8SUB              = OPR_SUB + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQSUB              = OPR_SUB + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_A4SUB              = OPR_SUB + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8SUB              = OPR_SUB + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10SUB             = OPR_SUB + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16SUB             = OPR_SUB + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16SUB             = OPR_SUB + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16SUB             = OPR_SUB + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_SWITCH             = OPR_SWITCH + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I1TAS              = OPR_TAS + RTYPE(MTYPE_I1) + DESC(MTYPE_V),
  OPC_I2TAS              = OPR_TAS + RTYPE(MTYPE_I2) + DESC(MTYPE_V),
  OPC_I4TAS              = OPR_TAS + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8TAS              = OPR_TAS + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U1TAS              = OPR_TAS + RTYPE(MTYPE_U1) + DESC(MTYPE_V),
  OPC_U2TAS              = OPR_TAS + RTYPE(MTYPE_U2) + DESC(MTYPE_V),
  OPC_U4TAS              = OPR_TAS + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8TAS              = OPR_TAS + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4TAS              = OPR_TAS + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8TAS              = OPR_TAS + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQTAS              = OPR_TAS + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10TAS             = OPR_TAS + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16TAS             = OPR_TAS + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4TAS              = OPR_TAS + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8TAS              = OPR_TAS + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQTAS              = OPR_TAS + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_A4TAS              = OPR_TAS + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8TAS              = OPR_TAS + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_C10TAS             = OPR_TAS + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16TAS             = OPR_TAS + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16TAS             = OPR_TAS + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16TAS             = OPR_TAS + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_TRAP               = OPR_TRAP + RTYPE(MTYPE_V) + DESC(MTYPE_V),

  OPC_I4TRIPLET          = OPR_TRIPLET + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8TRIPLET          = OPR_TRIPLET + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4TRIPLET          = OPR_TRIPLET + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8TRIPLET          = OPR_TRIPLET + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16TRIPLET         = OPR_TRIPLET + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16TRIPLET         = OPR_TRIPLET + RTYPE(MTYPE_U16) + DESC(MTYPE_V),

  OPC_TRUEBR             = OPR_TRUEBR + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4I4TRUNC          = OPR_TRUNC + RTYPE(MTYPE_I4) + DESC(MTYPE_I4),
  OPC_I4F4TRUNC          = OPR_TRUNC + RTYPE(MTYPE_I4) + DESC(MTYPE_F4),
  OPC_I4F8TRUNC          = OPR_TRUNC + RTYPE(MTYPE_I4) + DESC(MTYPE_F8),
  OPC_I4FQTRUNC          = OPR_TRUNC + RTYPE(MTYPE_I4) + DESC(MTYPE_FQ),
  OPC_I4F10TRUNC         = OPR_TRUNC + RTYPE(MTYPE_I4) + DESC(MTYPE_F10),
  OPC_I4F16TRUNC         = OPR_TRUNC + RTYPE(MTYPE_I4) + DESC(MTYPE_F16),
  OPC_I8F4TRUNC          = OPR_TRUNC + RTYPE(MTYPE_I8) + DESC(MTYPE_F4),
  OPC_I8F8TRUNC          = OPR_TRUNC + RTYPE(MTYPE_I8) + DESC(MTYPE_F8),
  OPC_I8FQTRUNC          = OPR_TRUNC + RTYPE(MTYPE_I8) + DESC(MTYPE_FQ),
  OPC_I8F10TRUNC         = OPR_TRUNC + RTYPE(MTYPE_I8) + DESC(MTYPE_F10),
  OPC_I8F16TRUNC         = OPR_TRUNC + RTYPE(MTYPE_I8) + DESC(MTYPE_F16),
  OPC_U4F4TRUNC          = OPR_TRUNC + RTYPE(MTYPE_U4) + DESC(MTYPE_F4),
  OPC_U4F8TRUNC          = OPR_TRUNC + RTYPE(MTYPE_U4) + DESC(MTYPE_F8),
  OPC_U4FQTRUNC          = OPR_TRUNC + RTYPE(MTYPE_U4) + DESC(MTYPE_FQ),
  OPC_U4F10TRUNC         = OPR_TRUNC + RTYPE(MTYPE_U4) + DESC(MTYPE_F10),
  OPC_U4F16TRUNC         = OPR_TRUNC + RTYPE(MTYPE_U4) + DESC(MTYPE_F16),
  OPC_U8F4TRUNC          = OPR_TRUNC + RTYPE(MTYPE_U8) + DESC(MTYPE_F4),
  OPC_U8F8TRUNC          = OPR_TRUNC + RTYPE(MTYPE_U8) + DESC(MTYPE_F8),
  OPC_U8FQTRUNC          = OPR_TRUNC + RTYPE(MTYPE_U8) + DESC(MTYPE_FQ),
  OPC_U8F10TRUNC         = OPR_TRUNC + RTYPE(MTYPE_U8) + DESC(MTYPE_F10),
  OPC_U8F16TRUNC         = OPR_TRUNC + RTYPE(MTYPE_U8) + DESC(MTYPE_F16),
  OPC_I16F4TRUNC         = OPR_TRUNC + RTYPE(MTYPE_I16) + DESC(MTYPE_F4),
  OPC_I16F8TRUNC         = OPR_TRUNC + RTYPE(MTYPE_I16) + DESC(MTYPE_F8),
  OPC_I16FQTRUNC         = OPR_TRUNC + RTYPE(MTYPE_I16) + DESC(MTYPE_FQ),
  OPC_I16F10TRUNC        = OPR_TRUNC + RTYPE(MTYPE_I16) + DESC(MTYPE_F10),
  OPC_I16F16TRUNC        = OPR_TRUNC + RTYPE(MTYPE_I16) + DESC(MTYPE_F16),
  OPC_U16F4TRUNC         = OPR_TRUNC + RTYPE(MTYPE_U16) + DESC(MTYPE_F4),
  OPC_U16F8TRUNC         = OPR_TRUNC + RTYPE(MTYPE_U16) + DESC(MTYPE_F8),
  OPC_U16FQTRUNC         = OPR_TRUNC + RTYPE(MTYPE_U16) + DESC(MTYPE_FQ),
  OPC_U16F10TRUNC        = OPR_TRUNC + RTYPE(MTYPE_U16) + DESC(MTYPE_F10),
  OPC_U16F16TRUNC        = OPR_TRUNC + RTYPE(MTYPE_U16) + DESC(MTYPE_F16),
  OPC_WHERE              = OPR_WHERE + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_WHILE_DO           = OPR_WHILE_DO + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_XGOTO              = OPR_XGOTO + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4XMPY             = OPR_XMPY + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8XMPY             = OPR_XMPY + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4XMPY             = OPR_XMPY + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8XMPY             = OPR_XMPY + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16XMPY            = OPR_XMPY + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16XMPY            = OPR_XMPY + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_XPRAGMA            = OPR_XPRAGMA + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_AFFIRM             = OPR_AFFIRM + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_A4ALLOCA           = OPR_ALLOCA + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8ALLOCA           = OPR_ALLOCA + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_U4ALLOCA           = OPR_ALLOCA + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8ALLOCA           = OPR_ALLOCA + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_DEALLOCA           = OPR_DEALLOCA + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_U4LDMA             = OPR_LDMA + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8LDMA             = OPR_LDMA + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_A4LDMA             = OPR_LDMA + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8LDMA             = OPR_LDMA + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_ASM_STMT           = OPR_ASM_STMT + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_ASM_INPUT          = OPR_ASM_INPUT + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4ASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8ASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4ASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8ASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4ASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8ASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10ASM_EXPR        = OPR_ASM_EXPR + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16ASM_EXPR        = OPR_ASM_EXPR + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4ASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8ASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_C10ASM_EXPR        = OPR_ASM_EXPR + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16ASM_EXPR        = OPR_ASM_EXPR + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16ASM_EXPR        = OPR_ASM_EXPR + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16ASM_EXPR        = OPR_ASM_EXPR + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_A4ASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8ASM_EXPR         = OPR_ASM_EXPR + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_U4U1RROTATE        = OPR_RROTATE + RTYPE(MTYPE_U4) + DESC(MTYPE_U1),
  OPC_U4U2RROTATE        = OPR_RROTATE + RTYPE(MTYPE_U4) + DESC(MTYPE_U2),
  OPC_U4U4RROTATE        = OPR_RROTATE + RTYPE(MTYPE_U4) + DESC(MTYPE_U4),
  OPC_U8U8RROTATE        = OPR_RROTATE + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPC_U4LDA_LABEL        = OPR_LDA_LABEL + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8LDA_LABEL        = OPR_LDA_LABEL + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_A4LDA_LABEL        = OPR_LDA_LABEL + RTYPE(MTYPE_A4) + DESC(MTYPE_V),
  OPC_A8LDA_LABEL        = OPR_LDA_LABEL + RTYPE(MTYPE_A8) + DESC(MTYPE_V),
  OPC_GOTO_OUTER_BLOCK   = OPR_GOTO_OUTER_BLOCK + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_I4EXTRACT_BITS     = OPR_EXTRACT_BITS + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8EXTRACT_BITS     = OPR_EXTRACT_BITS + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4EXTRACT_BITS     = OPR_EXTRACT_BITS + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8EXTRACT_BITS     = OPR_EXTRACT_BITS + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I4COMPOSE_BITS     = OPR_COMPOSE_BITS + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8COMPOSE_BITS     = OPR_COMPOSE_BITS + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4COMPOSE_BITS     = OPR_COMPOSE_BITS + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8COMPOSE_BITS     = OPR_COMPOSE_BITS + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  
  OPC_I1PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_I1),
  OPC_I2PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_I2),
  OPC_I4PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_I4),
  OPC_I8PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_I8),
  OPC_U1PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_U1),
  OPC_U2PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_U2),
  OPC_U4PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_U4),
  OPC_U8PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_U8),
  OPC_F4PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_F4),
  OPC_F8PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_F8),
  OPC_FQPSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_FQ),
  OPC_F10PSTORE          = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_F10),
  OPC_F16PSTORE          = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_F16),
  OPC_C4PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_C4),
  OPC_C8PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_C8),
  OPC_CQPSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_CQ),
  OPC_MPSTORE            = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_M),
  OPC_BSPSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_BS),
  OPC_A4PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_A4),
  OPC_A8PSTORE           = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_A8),
  OPC_C10PSTORE          = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_C10),
  OPC_C16PSTORE          = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_C16),
  OPC_I16PSTORE          = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_I16),
  OPC_U16PSTORE          = OPR_PSTORE + RTYPE(MTYPE_V) + DESC(MTYPE_U16),
  OPC_I1PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_I1),
  OPC_I2PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_I2),
  OPC_I4PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_I4),
  OPC_I8PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_I8),
  OPC_U1PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_U1),
  OPC_U2PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_U2),
  OPC_U4PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_U4),
  OPC_U8PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_U8),
  OPC_F4PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_F4),
  OPC_F8PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_F8),
  OPC_FQPSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_FQ),
  OPC_F10PSTID            = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_F10),
  OPC_F16PSTID            = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_F16),
  OPC_C4PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_C4),
  OPC_C8PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_C8),
  OPC_CQPSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_CQ),
  OPC_MPSTID              = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_M),
  OPC_BSPSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_BS),
  OPC_A4PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_A4),
  OPC_A8PSTID             = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_A8),
  OPC_C10PSTID            = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_C10),
  OPC_C16PSTID            = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_C16),
  OPC_I16PSTID            = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_I16),
  OPC_U16PSTID            = OPR_PSTID + RTYPE(MTYPE_V) + DESC(MTYPE_U16),

  OPC_VUSE           = OPR_USE + RTYPE(MTYPE_V) + DESC(MTYPE_V),
  OPC_BUSE           = OPR_USE + RTYPE(MTYPE_B) + DESC(MTYPE_V),
  OPC_NAMELIST       = OPR_NAMELIST+RTYPE(MTYPE_V)+DESC(MTYPE_V),
  OPC_IMPLICIT_BND   = OPR_IMPLICIT_BND+RTYPE(MTYPE_V)+DESC(MTYPE_V),

  OPC_I4SRCTRIPLET          = OPR_SRCTRIPLET + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8SRCTRIPLET          = OPR_SRCTRIPLET + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4SRCTRIPLET          = OPR_SRCTRIPLET + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8SRCTRIPLET          = OPR_SRCTRIPLET + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_I16SRCTRIPLET         = OPR_SRCTRIPLET + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16SRCTRIPLET         = OPR_SRCTRIPLET + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
 
  OPC_NULLIFY        = OPR_NULLIFY+RTYPE(MTYPE_V)+DESC(MTYPE_V),
  OPC_INTERFACE      = OPR_INTERFACE+RTYPE(MTYPE_V)+DESC(MTYPE_V),

  OPC_I4ARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8ARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4ARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8ARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4ARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8ARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10ARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16ARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4ARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8ARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQARRAY_CONSTRUCT   = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MARRAY_CONSTRUCT    = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_C10ARRAY_CONSTRUCT  = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16ARRAY_CONSTRUCT  = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16ARRAY_CONSTRUCT  = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16ARRAY_CONSTRUCT  = OPR_ARRAY_CONSTRUCT + RTYPE(MTYPE_U16) + DESC(MTYPE_V),

  OPC_I4IMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_I4) + DESC(MTYPE_V),
  OPC_I8IMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_I8) + DESC(MTYPE_V),
  OPC_U4IMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_U4) + DESC(MTYPE_V),
  OPC_U8IMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_U8) + DESC(MTYPE_V),
  OPC_F4IMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_F4) + DESC(MTYPE_V),
  OPC_F8IMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_F8) + DESC(MTYPE_V),
  OPC_FQIMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_FQ) + DESC(MTYPE_V),
  OPC_F10IMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_F10) + DESC(MTYPE_V),
  OPC_F16IMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_F16) + DESC(MTYPE_V),
  OPC_C4IMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_C4) + DESC(MTYPE_V),
  OPC_C8IMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_C8) + DESC(MTYPE_V),
  OPC_CQIMPLIED_DO   = OPR_IMPLIED_DO + RTYPE(MTYPE_CQ) + DESC(MTYPE_V),
  OPC_MIMPLIED_DO    = OPR_IMPLIED_DO + RTYPE(MTYPE_M) + DESC(MTYPE_V),
  OPC_C10IMPLIED_DO  = OPR_IMPLIED_DO + RTYPE(MTYPE_C10) + DESC(MTYPE_V),
  OPC_C16IMPLIED_DO  = OPR_IMPLIED_DO + RTYPE(MTYPE_C16) + DESC(MTYPE_V),
  OPC_I16IMPLIED_DO  = OPR_IMPLIED_DO + RTYPE(MTYPE_I16) + DESC(MTYPE_V),
  OPC_U16IMPLIED_DO  = OPR_IMPLIED_DO + RTYPE(MTYPE_U16) + DESC(MTYPE_V),
  OPC_U8U8STRCTFLD          = OPR_STRCTFLD + RTYPE(MTYPE_U8) + DESC(MTYPE_U8),
  OPCODE_LAST            = 0x3FFFF

} OPCODE;

/* typedef UINT32 OPCODE; */

/* operators, types, and attributes of the operators */
/* eraxxon (2005.01): Re-implement table and routines to support b2a
   and a2b conversions. */
struct OPERATOR_info_struct : public ir_a2b::enum2str_tbl_entry_t {

  OPERATOR_info_struct(OPERATOR opr_ = OPERATOR_UNKNOWN, 
		       const char* name_ = NULL,
		       INT8 nkids_ = 0,
		       OPERATOR_MAPCAT mapcat_ = OPERATOR_MAPCAT_HDR,
		       UINT32 flags_ = 0)
    : opr(opr_), _name(name_), nkids(nkids_), mapcat(mapcat_), _flags(flags_)
    { }
  virtual ~OPERATOR_info_struct() { }

  virtual INT getEnumVal() const     { return opr; }
  virtual const char* getStr() const { return _name; }

  OPERATOR        opr;
  const char*     _name;
  mINT8		  nkids;
  OPERATOR_MAPCAT mapcat;
  mUINT32	  _flags;
};

extern OPERATOR_info_struct OPERATOR_info[];

#undef RTYPE
#undef DESC

#ifdef __cplusplus
}
#endif

#endif /* ifndef opcode_gen_core_INCLUDED */
